# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose: weibo水军类垃圾检测
# author : 
# create_time : 2020/6/29 15:40
# update_time : 2020/6/29 15:40
# copyright : Lavector
# ----------------------------------------------
import re
import emoji
import codecs
import numpy as np
from lavda.analysis.da_clean.clean_interface import *
from lavda.util.mysql_util import *


class WaterRuleInterface(object):
    def predict(self, text, params=None):
        """
        :param text:
        :param params:
        :return:
        """
        pass


class WaterOrdinaryWays(object):
    """
    抽取的普通方法
    """

    @staticmethod
    def get_from2mysql(c_type="adword"):
        sql = """
        select keyword from %s
        """ % c_type
        # print(sql)
        result = np.array(LavwikiReader.execute_all(sql)[0])
        result = [i[0] for i in result]
        return result

    @staticmethod
    def get_clean(c_type):
        """
        type:mysql数据库表名，处理weibo垃圾、水军时的词

        """
        clean = WaterOrdinaryWays.get_from2mysql(c_type)
        clean = [str(i).lower().strip() for i in clean if len(str(i).strip()) > 0]
        return clean

    @staticmethod
    def read_by_line(file_path):
        """
        按行读取数据
        :param file_path:
        :return:
        """
        data_list = list()
        with codecs.open(file_path, 'r', encoding="utf-8") as fp:
            for line in fp:
                line = line.strip()
                if len(line) > 0:
                    data_list.append(line)

        return data_list

    @staticmethod
    def has_chinese_count(text):
        """
        判断中文字符数量
        :param self:
        :param text:
        :return:
        """
        count = 0
        tmp_text = text
        if isinstance(tmp_text, str):
            tmp_text = str(tmp_text)
        for ch in tmp_text:
            if '\u4e00' <= ch <= '\u9fff':
                count += 1
        return count

    @staticmethod
    def get_weibo_topic(text):
        tmp_text = text
        p1 = re.compile('#.*?#', re.S)  # 最小匹配
        result_list = re.findall(p1, tmp_text)
        return result_list

    @staticmethod
    def get_weibo_user_tag(text):
        r_user_end = re.compile('@[\S]+')
        result = r_user_end.findall(text)
        return result


class WaterWordRule(WaterRuleInterface):
    """
    根据是否包括特定词来判断
    这些广告词必须有很强的代表性，即当出现此广告词时，这条文本几乎肯定是广告
    """

    def __init__(self):
        self.word_set = set()
        self._load()

    def _load(self):
        # 1.加载广告词
        word_list = WaterOrdinaryWays.get_clean('weibo_water_word')
        for word in word_list:
            tmp_word = word.strip()
            if tmp_word and len(tmp_word) > 0:
                self.word_set.add(tmp_word)

    def predict(self, text, params=None):
        for word in self.word_set:
            if word in text:
                return True
        return False


class WaterHastagRule(WaterRuleInterface):
    """
    基于Hasttag判断是否为广告
    """
    def __init__(self):
        self.item_rule_list = list()
        self._load()

    def _load(self):
        # 2.加载广告item规则
        rule_list = WaterOrdinaryWays.get_clean('weibo_water_hastag_rule')
        tmp_rule_set = set()
        for rule in rule_list:
            tmp_rule = rule.strip()
            if tmp_rule:
                tmp_rule_set.add(tmp_rule)

        for tmp_rule in tmp_rule_set:
            tmp_rule_pattern = re.compile(tmp_rule, re.S)
            self.item_rule_list.append(tmp_rule_pattern)

    def _get_bracket(self, text, bracket, drop_bracket=True):
        """
        找出原文中带【】的字符串。result_list
        :param text:
        :param bracket:
        :param drop_bracket:
        :return:
        """
        tmp_text = text
        re_string = u"[{}].*?[{}]".format(bracket[0], bracket[1])
        p1 = re.compile(re_string, re.S)  # 最小匹配
        find_list = re.findall(p1, tmp_text)

        result_list = list()
        if find_list:
            for item in find_list:
                tmp_item = item
                if drop_bracket:
                    tmp_item = tmp_item.replace(bracket[0], u'')
                    tmp_item = tmp_item.replace(bracket[1], u'')
                result_list.append(tmp_item)
        return result_list

    def is_match_item(self, text, params=None):
        """
        原文把找出的词的【】去掉，和词库中的词比较，词在库中就返回true;
        :param text:
        :param params:
        :return:
        """
        for re_rule_pattern in self.item_rule_list:
            result = re_rule_pattern.search(text)
            if result and result.group() is not None:
                return True
        return False

    def predict(self, text, params=None):
        r2_list = self._get_bracket(text, (u'【', u'】'))
        bracket_list = r2_list

        for item in bracket_list:
            tmp_item = item.strip()
            if self.is_match_item(tmp_item):
                return True
        return False


class WaterRepeatPrefix(WaterRuleInterface):
    """
    根据连续出现的前缀进行判断，这些前缀可能就是明星名字
    """
    # 注意最后一个是空格
    SPLIT_CHAR = u'！， |  '
    MAX_REPEAT_COUNT = 5
    EXCEPTION_WORD = [u'比如']

    def predict(self, text, params=None):

        # 1.根据前缀进行分析
        topic_text_list = self._divide_by_hastag_topic(text)
        if self._has_prefix(topic_text_list):
            return True
        user_tag_text_list = self._divide_by_hastag_user_tag(text)
        if self._has_prefix(user_tag_text_list):
            return True
        punc_text_list = self._divide_by_punc(text)
        if self._has_prefix(punc_text_list):
            return True
        emoji_text_list = self._divide_by_emoji(text)
        if self._has_prefix(emoji_text_list):
            return True

        # 2. user tag和topic中出现同名
        user_tag_set = set(user_tag_text_list)
        topic_set = set(topic_text_list)
        for user in user_tag_set:
            for topic in topic_set:
                if user in topic:
                    return True

        return False

    def _has_prefix(self, text_list, ngram=True):
        if ngram:
            max_count = self._get_max_ngram(text_list, 2)
            if max_count >= self.MAX_REPEAT_COUNT:
                return True
            max_count = self._get_max_ngram(text_list, 3)
            if max_count >= self.MAX_REPEAT_COUNT:
                return True
        else:
            max_count = 0
            gram_dict = dict()
            for item in text_list:
                if item in gram_dict:
                    gram_dict[item] += 1
                else:
                    gram_dict[item] = 1
                if gram_dict[item] > max_count:
                    max_count = gram_dict[item]

            if max_count >= self.MAX_REPEAT_COUNT:
                return True

        return False

    def _get_max_ngram(self, text_list, n):
        ngram_dict = dict()
        max_count = 0
        for item in text_list:
            sub_item = item[:n]
            if sub_item in self.EXCEPTION_WORD:
                continue
            # TODO 止前gram必须全是汉字
            if len(sub_item) != n:
                continue
            if WaterOrdinaryWays.has_chinese_count(sub_item) != n:
                continue
            if not sub_item:
                continue
            if sub_item in ngram_dict:
                ngram_dict[sub_item] += 1
            else:
                ngram_dict[sub_item] = 1
            if ngram_dict[sub_item] > max_count:
                max_count = ngram_dict[sub_item]

        return max_count

    def _divide_by_hastag_topic(self, text):
        tmp_text_list = WaterOrdinaryWays.get_weibo_topic(text)
        text_list = list()
        for item in tmp_text_list:
            item = item.replace(u'#', u'').strip()
            if not item:
                continue
            text_list.append(item)

        return text_list

    def _divide_by_hastag_user_tag(self, text):
        tmp_text_list = WaterOrdinaryWays.get_weibo_user_tag(text)
        text_list = list()
        for item in tmp_text_list:
            item = item.replace(u'@', u'').strip()
            if not item:
                continue
            text_list.append(item)

        return text_list

    def _divide_by_punc(self, text):
        """
        根据标点符号切分
        :param text:
        :return:
        """
        re_string = u'[{}]'.format(self.SPLIT_CHAR)
        text_list = re.split(re_string, text)
        text_list = [item.strip() for item in text_list if item.strip()]

        return text_list

    def _divide_by_emoji(self, text):
        """
        根据表情符号切分
        :param text:
        :return:
        """
        emoji_text = emoji.demojize(text)

        # 找到所有表情符
        p1 = re.compile(':.*?:', re.S)  # 最小匹配
        result_list = re.findall(p1, emoji_text)
        emoji_set = set()
        for item in result_list:
            item = item.replace(u':', u'')
            emoji_set.add(item)

        # 按表情拆分
        sub_text_list = re.split('[:.*?:]', emoji_text)

        # 去掉表情符
        text_list = list()
        for item in sub_text_list:
            if item in emoji_set:
                continue
            text_list.append(item)

        return text_list


class WaterForwardMicroblogRule(WaterRuleInterface):
    """
    以  转发微博  或 report  开头的都算水军
    明星名字出现6次以上都算水军
    """
    def __init__(self):
        sql = "select name from lav2bgc_star where channel='weibo' and classify='personal';"
        self.weibo_star_list = [i[0] for i in np.array(LavwikiReader.execute_all(sql)[0])]
    def predict(self, text, params=None):
        if re.match(r"^转发微博", text) or re.match(r"^report", text.lower()):
            return True
        for i in self.weibo_star_list:
            text_list = text.split(i)
            text_list_len = len(text_list)
            if text_list_len > 6:
                return True
        return False


class WeiboWaterRuleModel(RowInterface, CleanInterface):
    """
    将以上三种规则整合在一起判断
    """
    def __init__(self, **params):
        super(WeiboWaterRuleModel, self).__init__(CleanFlag.NO)

        self.rule_list = list()
        self.rule_list.append(WaterWordRule())
        self.rule_list.append(WaterRepeatPrefix())
        self.rule_list.append(WaterHastagRule())
        self.rule_list.append(WaterForwardMicroblogRule())

    def process(self, data, **params):
        clean_text = str(data).lower()
        for rule in self.rule_list:
            if rule.predict(clean_text):
                return CleanFlag.YES
        return CleanFlag.NO


def main():

    sub_model_water = WeiboWaterRuleModel()

    # content = u'【19.9】宝丽康美 核桃黑芝麻黑豆代餐粉'
    content = u'活粉@助力,@助力,@助力,@助力,@助力,@助力,@助力,@助力,@助力4hd44,'
    channel = u"weibo"
    # df = PandaTool.read(r'C:\Users\zhang\Desktop\test.xlsx')

    # result = df.apply(lambda line: sub_model_water.process(line["content"], line["channel"]), axis=1).astype(int)
    result = sub_model_water.process(content)
    print(result)


if __name__ == '__main__':
    main()



# -*- coding:utf-8 -*-
"""
DataFrame 文件数据合并相差功能
"""
import os
import pandas as pd

from lavda.backexcutor.analysis_interface import *
from lavda.analysis.da_pre.create_data_index import *
from lavda.util.panda_tool import PandaTool
from lavda.util.mysql_util import LavwikiDfReader


class DfFormatData(DfMergeInterface):
    """
    根据channel将不同数据合并，同时设置索引
    """
    def __init__(self, index_column="", channel_column="channel"):
        self.index_column = index_column
        if channel_column:
            self.channel_column = channel_column
        else:
            self.channel_column = None
        self.merge_dict = dict()
        self.merge_st_names = list()

        # 从数据库加载映射表
        self.load_from_mysql()

    def load_from_mysql(self):
        """
        从数据库加载映射表
        :return:
        """
        df_input = LavwikiDfReader.read_table("config_merge")

        self._parse_from_df(df_input)

    def load_config_data(self, **config):
        """
        从文件加载映射表
        :param config:
        :return:
        """
        if not config:
            return
        if "mergeDictPath" not in config:
            return

        merge_dict_path = config['mergeDictPath']
        # TODO 输入的资源文件是可选的，主要还是使用数据库文件
        if not os.path.exists(merge_dict_path):
            return

        df_input = PandaTool.read(merge_dict_path)
        self._parse_from_df(df_input)

    def _parse_from_df(self, df_input):
        df_input = PandaTool.drop_unnamed_column(df_input)
        column_list = PandaTool.get_column(df_input)

        # TODO 第1列表示channel，其他列为各数据源字段
        channel_column = column_list[0]
        other_column = column_list[1:]
        result_dict = dict()
        for index, row in df_input.iterrows():
            channel_name = row[channel_column]
            if pd.isna(channel_name):
                # raise Exception("channel name is none, index =" + str(index))
                continue
            format_channel_name = channel_name

            result_dict[format_channel_name] = dict()

            for column_name in other_column:
                field_name = row[column_name]
                if pd.isna(field_name):
                    continue

                field_name = str(field_name).strip()
                if not field_name:
                    continue

                # TODO 对齐后标准名要去空格
                result_dict[format_channel_name][field_name] = str(column_name).strip()

        # 合并结果数据
        self.merge_dict.update(result_dict)
        for col in other_column:
            if col in self.merge_st_names:
                continue
            self.merge_st_names.append(col)

    def process(self, input_path, **params):
        """
        
        :param input_path: list(file_path), 必须是完整路径
        :return: 
        """
        file_params_list = input_path

        df_list = list()
        may_data_count = 0
        for file_path in file_params_list:
            parent_file_path, temp_file_name = os.path.split(file_path)
            file_name, extension = os.path.splitext(temp_file_name)

            channel = file_name.split("_")[1].strip().lower()
            # TODO 没有映射词表的不处理，不合并
            if channel not in self.merge_dict:
                raise Exception("can not merge data for channel = " + str(channel))
            # 保留标签类数据
            one_channel_dict = self.merge_dict[channel]
            df_input = PandaTool.read(file_path)
            # print("sub file path = ", file_path)
            # print("sub file shape = ", df_input.shape)
            may_data_count += df_input.shape[0]

            all_df_column = PandaTool.get_column(df_input)
            exist_map_dict = dict()
            not_exist_column = list()
            for col in all_df_column:
                if col in one_channel_dict:
                    exist_map_dict[col] = one_channel_dict[col]
                elif col in self.merge_st_names:
                    # TODO 已经是标准化的名字，也添加到里面，以保留
                    exist_map_dict[col] = col
                else:
                    not_exist_column.append(col)

            df_input.rename(columns=exist_map_dict, inplace=True)

            # TODO 筛选指定的数据列，不在词典中的列一律删掉
            df_input = df_input[list(exist_map_dict.values())]
            df_list.append(df_input)

            # 填充channel
            if self.channel_column:
                df_input[self.channel_column] = channel

        df_result = pd.concat(df_list, sort=False)
        # print("label result shape:", df_result.shape)
        real_data_count = df_result.shape[0]
        if real_data_count != may_data_count:
            raise Exception("data merge error, last total count error")

        # TODO 对于存在于词表中，但不存在于数据中的列，填充空数据
        new_column = PandaTool.get_column(df_result)
        fill_columns = [col for col in self.merge_st_names if col not in new_column]
        for column in fill_columns:
            df_result[column] = None

        # TODO 生成数据id
        data_index_object = CreateDataIndex()
        if self.index_column:
            df_result[self.index_column] = [data_index_object.process() for _ in range(real_data_count)]

        df_result = df_result[self.merge_st_names]

        return df_result


if __name__ == "__main__":
    DfFormatData()
    DfFormatData.load_config_data(None)



# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose: 城市等级数据提取
# author : 
# create_time : 2020/6/29 15:01
# update_time : 2020/6/29 15:01
# copyright : Lavector
# ----------------------------------------------
import pandas as pd

from lavda.util.panda_tool import PandaTool
from lavda.util.mysql_util import LavwikiReader
from lavda.backexcutor.analysis_interface import RowInterface
from lavda.analysis.da_extract.area import *


class AreaDictTool(object):
    def _load_main_area_from_db(self, mysql_table_name, integer_column):
        """
        从数据库加载词典信息
        :return:
        """
        df_input = LavwikiReader.get_as_df(mysql_table_name)

        column_list = PandaTool.get_column(df_input)
        column_list = [column for column in column_list if "unname" not in column]

        area_list = list()
        for index, row in df_input.iterrows():
            area_item = dict()
            for column in column_list:
                if column in integer_column:
                    area_item[column] = int(row[column])
                else:
                    area_item[column] = str(row[column])

            area_list.append(area_item)

        return area_list

    def _load_from_excel(self, input_file_path, integer_column):
        df_input = PandaTool.read(input_file_path)

        column_list = PandaTool.get_column(df_input)
        column_list = [column for column in column_list if "unname" not in column]

        area_list = list()
        for index, row in df_input.iterrows():
            area_item = dict()
            for column in column_list:
                if column in integer_column:
                    area_item[column] = int(row[column])
                else:
                    area_item[column] = str(row[column])

            area_list.append(area_item)

        return area_list

    def _save_to_excel(self, area_list, integer_column, result_file_path):
        columns = list(area_list[0].keys())

        result_list = list()
        for item in area_list:
            item_list = [item[column] for column in columns]
            result_list.append(item_list)

        df_result = pd.DataFrame(result_list, columns=columns)
        for column in integer_column:
            df_result[column] = df_result[column].astype("int")

        PandaTool.save(result_file_path, df_result)

    def load_main_area_from_excel(self, input_file_path):
        """
        适用于各种词典
        :param input_file_path:
        :return:
        """
        integer_column = {'id', "level"}

        area_list = self._load_from_excel(input_file_path, integer_column)
        # area_list = self._load_main_area_from_db(mysql_table_name, integer_column)
        dict_object = MainAreaDict(area_list)

        return dict_object

    def load_main_area_from_db(self, mysql_table_name):
        """
        适用于各种词典
        :param input_file_path:
        :return:
        """
        integer_column = {'id', "level"}

        area_list = self._load_main_area_from_db(mysql_table_name, integer_column)
        dict_object = MainAreaDict(area_list)

        return dict_object

    def save_main_area_to_excel(self, main_area_dict_object, result_file_path):
        """
        将词典保存为excel文件
        :param main_area_dict_object:
        :param result_file_path:
        :return:
        """
        area_list = main_area_dict_object.get()
        self._save_to_excel(area_list, ['id', 'level'], result_file_path)

    def load_city_rank_from_excel(self, input_file_path):
        """
        适用于各种词典
        :param input_file_path:
        :return:
        """
        integer_column = {'id', 'rankLevel', "level"}
        area_list = self._load_from_excel(input_file_path, integer_column)

        dict_object = CityRankDict(area_list)

        return dict_object

    def save_city_rank_to_excel(self, city_rank_dict_object, result_file_path):
        area_list = city_rank_dict_object.get()
        self._save_to_excel(area_list, ['id', 'level', 'rankLevel'], result_file_path)

    def load_custom_city_rank_from_excel(self, input_file_path,
                                         sheet_name=None,
                                         city_column='city',
                                         rank_column='cityRank'):
        """
        加载自定义城市级别词典
        :param input_file_path: string,文件路径
        :param city_column: string, 表示城市的列名
        :param rank_column: string, 表示城市级别的列名
        :return:
        """
        if sheet_name:
            df_input = PandaTool.read_more(input_file_path, sheet_name=sheet_name)
        else:
            df_input = PandaTool.read_more(input_file_path)

        custom_rank_dict = dict()
        for index, row in df_input.iterrows():
            city_name = row[city_column]
            city_rank = row[rank_column]

            if pd.isna(city_name) or pd.isna(city_rank):
                continue
            city_name = str(city_name).strip()
            city_rank = city_rank

            custom_rank_dict[city_name] = city_rank

        return custom_rank_dict

    def load_custom_city_rank_from_db(self, mysql_table_name,
                                         sheet_name=None,
                                         city_column='city',
                                         rank_column='cityRank'):
        """
        加载自定义城市级别词典
        :param input_file_path: string,文件路径
        :param city_column: string, 表示城市的列名
        :param rank_column: string, 表示城市级别的列名
        :return:
        """
        df_input = LavwikiReader.get_as_df(mysql_table_name)

        custom_rank_dict = dict()
        for index, row in df_input.iterrows():
            city_name = row[city_column]
            city_rank = row[rank_column]

            if pd.isna(city_name) or pd.isna(city_rank):
                continue
            city_name = str(city_name).strip()
            city_rank = city_rank

            custom_rank_dict[city_name] = city_rank

        return custom_rank_dict


def gl_city_rank_filter(find_city_item):
    """
    城市级别查找时，增加一些过虑规则
    :return:
    """
    if len(find_city_item) > 1:
        # 1.1 检查并缩小范围
        tmp_dict_count_dict = dict()
        for item in find_city_item:
            current_level = item['level']
            if current_level not in tmp_dict_count_dict:
                tmp_dict_count_dict[current_level] = 0
            tmp_dict_count_dict[current_level] += 1

        if 1 in tmp_dict_count_dict and 2 in tmp_dict_count_dict:
            # TODO 同时存在1级、2级行政区域的，无法处理，
            return None
        if 1 in tmp_dict_count_dict and 3 in tmp_dict_count_dict:
            # TODO 同时存在1级、2级行政区域的，无法处理，
            return None
        if 2 in tmp_dict_count_dict and 3 in tmp_dict_count_dict:
            # TODO 同时存在2级、3级行政区域的，挑选第2级
            find_city_item = [item for item in find_city_item if item['level'] == 2]

    return find_city_item


class CustomCityRank(RowInterface):
    """
    根据自定义城市级别信息，提取地级城市、直辖市，然后分级、分类
    """

    def __init__(self,
                 default_value=None):
        self.default_value = default_value, default_value
        super(CustomCityRank, self).__init__(self.default_value)

        dict_tool = AreaDictTool()
        custom_rank_dict = dict_tool.load_custom_city_rank_from_db('lav2_custom_city_rank')
        self.main_area_object = dict_tool.load_main_area_from_db("lav2_china_area")

        # 20210819新加的，判断城市用
        self.city_list = list(custom_rank_dict.keys())

        self.custom_rank_detail_dict = dict()

        extract_processor = CityExtractProcessor(self.main_area_object, self.city_list)
        for name in custom_rank_dict:
            # TODO 应该只有1、2级城市，不应该会有县级城市，一级只会是直辖市，不会有省名
            find_city_item = extract_processor.process(name)
            if find_city_item is None or len(find_city_item) <= 0:
                # print("unknown city------==", str(name))
                continue
            if len(find_city_item) > 1:
                # TODO 有多个行政区域，且重名的，挑选2级区域，一级只会是直辖市，不会有省名
                find_city_item = [item for item in find_city_item if item['level'] == 2]
            if len(find_city_item) != 1:
                # print(find_city_item)
                # print("can not find city, may have many, name =", name)
                continue
            find_city_item = find_city_item[0]

            item = dict()
            item['areaCode'] = find_city_item['areaCode']
            item['shortName'] = find_city_item['shortName']
            item['level'] = find_city_item['level']
            item['name'] = find_city_item['name']
            item['oldName'] = name
            item['rankLevel'] = custom_rank_dict[name]

            self.custom_rank_detail_dict[item['areaCode']] = item

    def process(self, data, params=None):
        """

        :param data: string
        :param params:
        :return:
        """
        extract_processor = CityExtractProcessor(self.main_area_object, self.city_list)
        find_city_item = extract_processor.process(data)

        default_result = self.default_value

        # 1.查找城市、并确定范围
        if find_city_item is None or len(find_city_item) <= 0:
            return default_result

        find_city_item = gl_city_rank_filter(find_city_item)
        if find_city_item is None:
            return default_result

        target_rank_item = None
        for city_item in find_city_item:
            # 1、2级城市直接返回信息信息
            city_area_code = city_item['areaCode']
            if city_area_code in self.custom_rank_detail_dict:
                target_rank_item = self.custom_rank_detail_dict[city_area_code]
                break
            # 3级城市返回它的父类
            parent_area_code = city_item['parent']
            if parent_area_code in self.custom_rank_detail_dict:
                target_rank_item = self.custom_rank_detail_dict[parent_area_code]
                break

        if target_rank_item is None:
            return default_result

        result = dict()
        result['city'] = str(target_rank_item['shortName'])
        result['city_rank'] = str(target_rank_item["rankLevel"])

        return result['city'], result['city_rank']


if __name__ == '__main__':
    city_rank = CustomCityRank()
    print(city_rank.process("重庆 铜梁县"))


# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose : 
# author : 
# create_time : 2020/7/15 16:28
# update_time : 2020/7/15 16:28
# copyright : Lavector
# ----------------------------------------------


def parse_multi_column(field_str):
    """
    解析字段名，由于配置文件是字符串，因此要转换成需要的list格式
    :param field_str: string
    :return: list(string)
    """
    if field_str is None:
        return list()
    column_list = field_str.split(",")
    column_list = [item.strip() for item in column_list if item.strip()]

    return column_list



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date  : 2021/3/1 2:40 下午
# @File  : label_model.py
# @Author: johnson
# @Contact : github: johnson7788
# @Desc  : memnet 模型预测推理
# 依赖包 pip install transformers==3.0.2 torch numpy
import os
import torch
import math
import torch.nn as nn
import numpy as np
import torch.nn.functional as F


class Attention(nn.Module):
    def __init__(self, embed_dim, hidden_dim=None, out_dim=None, n_head=1, score_function='dot_product', dropout=0):
        ''' Attention Mechanism 注意力机制
        :param embed_dim: 嵌入维度; eg: 768
        :param hidden_dim: 隐藏维度; eg: 96
        :param out_dim: 输出维度   eg:300
        :param n_head: num of head (Multi-Head Attention)  head的数量  eg:8
        :param score_function: scaled_dot_product / mlp (concat) / bi_linear (general dot)   eg: mlp， bi_linear， dot_product, scaled_dot_product
        :return (?, q_len, out_dim,)
        '''
        super(Attention, self).__init__()
        if hidden_dim is None:
            hidden_dim = embed_dim // n_head
        if out_dim is None:
            out_dim = embed_dim
        self.embed_dim = embed_dim
        self.hidden_dim = hidden_dim
        self.n_head = n_head
        self.score_function = score_function
        # Q和K的计算
        self.w_k = nn.Linear(embed_dim, n_head * hidden_dim)
        self.w_q = nn.Linear(embed_dim, n_head * hidden_dim)
        # attention的投影层计算
        self.proj = nn.Linear(n_head * hidden_dim, out_dim)
        self.dropout = nn.Dropout(dropout)
        if score_function == 'mlp':
            self.weight = nn.Parameter(torch.Tensor(hidden_dim * 2))
        elif self.score_function == 'bi_linear':
            self.weight = nn.Parameter(torch.Tensor(hidden_dim, hidden_dim))
        else:  # dot_product / scaled_dot_product
            self.register_parameter('weight', None)
        self.reset_parameters()

    def reset_parameters(self):
        """权重初始化"""
        stdv = 1. / math.sqrt(self.hidden_dim)
        if self.weight is not None:
            self.weight.data.uniform_(-stdv, stdv)

    def forward(self, k, q):
        """
        计算MHA， Mutil Head attention
        :param k: 默认输入维度 [batch_size, seq_length, embedding_dim]
        :param q:  [batch_size, seq_length, embedding_dim]
        :return:
        """
        if len(q.shape) == 2:  # q_len missing
            q = torch.unsqueeze(q, dim=1)
        if len(k.shape) == 2:  # k_len missing
            k = torch.unsqueeze(k, dim=1)
        # mb_size 是batch_size
        mb_size = k.shape[0]  # ?
        # k_len和q_len是seq_len
        k_len = k.shape[1]
        q_len = q.shape[1]
        # score: (n_head*?, q_len, k_len,)
        # output: (?, q_len, out_dim,)
        # self.w_k(k)的维度[batch-size,seq_len, n_head * hidden_dim]
        # kx 维度 [batch-size,seq_len, n_head, hidden_dim]
        kx = self.w_k(k).view(mb_size, k_len, self.n_head, self.hidden_dim)
        # 【heads, batch_size, seq_len, hidden_dim】kx.permute(2, 0, 1, 3).contiguous().shape --> torch.Size([8, 16, 55, 96])
        # view(-1, k_len, self.hidden_dim) 合并head和bath_size维度 [heads * batch_size, seq_len,hidden_dim]
        kx = kx.permute(2, 0, 1, 3).contiguous().view(-1, k_len, self.hidden_dim)
        # qx同 kx维度
        qx = self.w_q(q).view(mb_size, q_len, self.n_head, self.hidden_dim)
        qx = qx.permute(2, 0, 1, 3).contiguous().view(-1, q_len, self.hidden_dim)
        #
        if self.score_function == 'dot_product':
            kt = kx.permute(0, 2, 1)
            score = torch.bmm(qx, kt)
        elif self.score_function == 'scaled_dot_product':
            kt = kx.permute(0, 2, 1)
            qkt = torch.bmm(qx, kt)
            score = torch.div(qkt, math.sqrt(self.hidden_dim))
        elif self.score_function == 'mlp':
            # torch.unsqueeze(kx, dim=1) 对kx扩充一个维度, [heads * batch_size, seq_len,hidden_dim]--> [heads * batch_size, 1, seq_len,hidden_dim]
            # expand(-1, q_len, -1, -1)， 把维度2扩充到seq_len维度, [heads * batch_size, seq_len, seq_len,hidden_dim]
            kxx = torch.unsqueeze(kx, dim=1).expand(-1, q_len, -1, -1)
            # torch.unsqueeze(qx, dim=2) -->  [heads * batch_size, seq_len,hidden_dim]  --> [heads * batch_size,seq_len, 1 ,hidden_dim]
            # 把维度3扩充到seq_len维度, [heads * batch_size, seq_len, seq_len,hidden_dim]
            qxx = torch.unsqueeze(qx, dim=2).expand(-1, -1, k_len, -1)
            # kq.shape (n_head*batch_size, q_len, k_len, hidden_dim*2)
            kq = torch.cat((kxx, qxx), dim=-1)
            # kq = torch.unsqueeze(kx, dim=1) + torch.unsqueeze(qx, dim=2)
            # score shape [heads * batch_size, q_len, seq_len]
            score = torch.tanh(torch.matmul(kq, self.weight))
        elif self.score_function == 'bi_linear':
            qw = torch.matmul(qx, self.weight)
            kt = kx.permute(0, 2, 1)
            score = torch.bmm(qw, kt)
        else:
            raise RuntimeError('invalid score_function')
        # score shape  [heads * batch_size, q_len, seq_len]
        score = F.softmax(score, dim=-1)
        # bmm， 批次的batch1和batch2内的矩阵进行批矩阵乘操作 [heads * batch_size, q_len, hidden_dim]
        output = torch.bmm(score, kx)  # (n_head*?, q_len, hidden_dim)
        # torch.split(output, mb_size, dim=0) --> 拆分出8个head，每个维度[batch_size, q_len,hidden_dim], 然后拼接所有n_head
        # output的维度 [batch_size, q_len,hidden_dim *n_head]
        output = torch.cat(torch.split(output, mb_size, dim=0), dim=-1)
        # [batch_size, seq_len, proj_layer_hidden_dim] eg: torch.Size([16, 55, 300])
        output = self.proj(output)  # (?, q_len, out_dim)
        output = self.dropout(output)
        return output, score


class SqueezeEmbedding(nn.Module):
    """
    挤压序列嵌入长度为批次中最长的
    由默认全部一样的默认序列长度，挤压成这个batch中最大的序列长度
    """

    def __init__(self, batch_first=True):
        """
        初始化
        :param batch_first:
        """
        super(SqueezeEmbedding, self).__init__()
        self.batch_first = batch_first

    def forward(self, x, x_len):
        """
        sequence -> sort -> pad and pack -> unpack ->unsort
        由默认全部一样的默认序列长度，挤压成这个batch中最大的序列长度,
        :param x: sequence embedding vectors  未排序的x [batch_size, seq_length]  eg: [16,80]
        :param x_len: numpy/tensor list   x的每个句子的长度 eg:[55,33,30,29,...]
        :return:  x维度 [16,55], 把多余的padding的0挤压出去了
        """
        """排序，从句子长度最长到最短"""
        x_sort_idx = torch.sort(-x_len)[1].long()
        x_unsort_idx = torch.sort(x_sort_idx)[1].long()
        # 长度从大到小排列
        x_len = x_len[x_sort_idx]
        # x也调整成这个排列
        x = x[x_sort_idx]
        """pack 将一个填充过的变长序列 压紧"""
        x_emb_p = torch.nn.utils.rnn.pack_padded_sequence(x, x_len.cpu(), batch_first=self.batch_first)
        """unpack: out, 这个操作和pack_padded_sequence()是相反的。把压紧的序列再填充回来"""
        out = torch.nn.utils.rnn.pad_packed_sequence(x_emb_p, batch_first=self.batch_first)  # (sequence, lengths)
        # unpack回来的序列x, out.shape [batch_size, max_seq_len]
        out = out[0]  #
        """unsort"""
        out = out[x_unsort_idx]
        return out


class Tokenizer(object):
    def __init__(self, max_seq_len, lower=True):
        """
        初始化tokenizer
        :param max_seq_len:  最大序列长度
        :param lower:  是否转换成小写
        """
        self.lower = lower
        self.max_seq_len = max_seq_len
        self.word2idx = {}
        self.idx2word = {}
        self.idx = 1

    def pad_and_truncate(self, sequence, maxlen, dtype='int64', padding='post', truncating='post', value=0):
        """
        对序列进行padding和截取
        :param sequence: 序列id，例如 [2021, 1996, 3095, 2001, 2061, 9202, 2000, 2149, 1012]
        :param maxlen: 最大序列长度
        :param dtype: 数据转换成numpy int64格式
        :param padding: post 还是pre
        :param truncating: post 还是pre
        :param value: 默认填充的值，默认用0填充
        :return: [2021 1996 3095 2001 2061 9202 2000 2149 1012    0    0    0    0    0,    0    0    0    0    0    0    0    0    0    0    0    0    0    0,    0    0    0    0    0    0    0    0    0    0    0    0    0    0,    0    0    0    0    0    0    0    0    0    0    0    0    0    0,    0    0    0    0    0    0    0    0    0    0    0    0    0    0,    0    0    0    0    0    0    0    0    0    0]
        """
        # 初始化创建maxlen序列长度的numpy 列表，
        x = (np.ones(maxlen) * value).astype(dtype)
        if truncating == 'pre':
            # 从前面截断
            trunc = sequence[-maxlen:]
        else:
            # 从后面截断
            trunc = sequence[:maxlen]
        # 截断后的数据转换成numpy
        trunc = np.asarray(trunc, dtype=dtype)
        # 从前还是后面padding
        if padding == 'post':
            x[:len(trunc)] = trunc
        else:
            x[-len(trunc):] = trunc
        return x

    def text_to_sequence(self, text, reverse=False, padding='post', truncating='post', english=False, max_seq_len=70):
        if self.lower:
            text = text.lower()
        if english:
            words = text.split()
        else:
            words = [w for w in text]
        unknownidx = len(self.word2idx) + 1
        sequence = [self.word2idx[w] if w in self.word2idx else unknownidx for w in words]
        if len(sequence) == 0:
            sequence = [0]
        if reverse:
            sequence = sequence[::-1]
        seq = self.pad_and_truncate(sequence, max_seq_len, padding=padding, truncating=truncating)
        return seq


class MemNet(nn.Module):

    def __init__(self, embedding_matrix, device, embed_dim=300, class_num=2):
        self.device = device
        #attention的迭代次数
        self.hops = 3
        super(MemNet, self).__init__()
        self.embed = nn.Embedding.from_pretrained(embedding_matrix, freeze=True)
        self.squeeze_embedding = SqueezeEmbedding(batch_first=True)
        self.attention = Attention(embed_dim, score_function='mlp')
        self.x_linear = nn.Linear(embed_dim, embed_dim)
        self.dense = nn.Linear(embed_dim, class_num)

    def forward(self, inputs):
        text_raw_without_keword_indices, keword_indices = inputs[0], inputs[1]
        memory_len = torch.sum(text_raw_without_keword_indices != 0, dim=-1)
        keword_len = torch.sum(keword_indices != 0, dim=-1)
        nonzeros_keword = keword_len.to(self.device)

        memory = self.embed(text_raw_without_keword_indices)
        memory = self.squeeze_embedding(memory, memory_len)
        keword = self.embed(keword_indices)
        keword = torch.sum(keword, dim=1)
        tmp_data = nonzeros_keword.view(nonzeros_keword.size(0), 1).to(torch.float)
        keword = torch.div(keword, tmp_data)
        x = keword.unsqueeze(dim=1)
        for _ in range(self.hops):
            x = self.x_linear(x)
            out_at, _ = self.attention(memory, x)
            x = out_at + x
        x = x.view(x.size(0), -1)
        out = self.dense(x)
        return out


class LabelComponentModel:
    """
    成分判别模型
    """
    def __init__(self, root_model_path='/data/share/nlp/label/component'):
        # 模型推理
        # 1代表是成分，0代表不是成分
        self.label_list = [0, 1]
        self.label_name = ["不是成分", "是成分"]
        # 模型状态文件
        self.state_dict_path = os.path.join(root_model_path, 'memnet_components_step1400_val_acc0.8947')
        # tokenizer文件
        self.tokenizer_file = os.path.join(root_model_path, 'components_tokenizer.dat')
        # embedding文件
        self.embeeding_file = os.path.join(root_model_path, '300_components_embedding_matrix.dat')
        self.embed_dim = 300
        self.hidden_dim = 300
        self.max_seq_len = 70
        self.class_num = len(self.label_list)
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.load_eval_model()

    def load_eval_model(self):
        self.tokenizer = self.load_tokenizer(dat_fname=self.tokenizer_file)
        embedding_matrix = self.load_embedding_matrix(dat_fname=self.embeeding_file)
        self.model = MemNet(embedding_matrix, device=self.device, embed_dim=self.embed_dim, class_num=self.class_num)
        print('加载模型:{0}'.format(self.state_dict_path))
        self.model.load_state_dict(torch.load(self.state_dict_path, map_location=self.device))
        self.model = self.model.to(self.device)
        # 模型评估
        self.model.eval()
        torch.autograd.set_grad_enabled(False)

    def load_embedding_matrix(self, dat_fname):
        """
        :param dat_fname:  缓存文件名字
        :return: embedding_matrix
        """
        print('加载 embedding_matrix:', dat_fname)
        embedding_matrix = torch.load(dat_fname, map_location=self.device)
        return embedding_matrix

    def load_tokenizer(self, dat_fname):
        """
        加载缓存好的tokeniner file
        :param dat_fname:  token 数据文件的名字 eg： 'restaurant_tokenizer.dat'
        :return:
        """
        print('加载 tokenizer 文件:', dat_fname)
        tokenizer = Tokenizer(self.max_seq_len)
        tokenizer.word2idx, tokenizer.idx2word = torch.load(dat_fname, map_location=self.device)
        return tokenizer

    def predict(self, content, keword, start, end):
        """
        单条数据
        :param content: 内容
        :param keword: 成分关键字
        :param start: 成分关键字开始位置索引
        :param end: 成分关键字结束位置索引
        :return:
        """
        if content[start:end] != keword:
            print("请注意keword的关键字的位置信息不准确")
        context_seq = self.tokenizer.text_to_sequence(content,max_seq_len=self.max_seq_len)
        keword_seq = self.tokenizer.text_to_sequence(keword,max_seq_len=self.max_seq_len)
        context_indices = torch.tensor([context_seq], dtype=torch.int64).to(self.device)
        keword_indices = torch.tensor([keword_seq], dtype=torch.int64).to(self.device)

        t_inputs = [context_indices, keword_indices]
        t_outputs = self.model(t_inputs)
        t_probs = F.softmax(t_outputs, dim=-1).cpu().detach().numpy()
        label_idx = t_probs.argmax(axis=-1)
        label_num = self.label_list[label_idx[0]]
        label_name = self.label_name[label_idx[0]]

        if label_num == 1:
            return True
        else:
            return False

    def predict_batch(self, test_data):
        """
        test_data数据格式 [[content,keword, start,end],...]
        eg: [['它含有的茶多酚能有效的阻止一些致癌物质的合成','茶多酚', 4,7],...]
        """
        context_seqs = []
        keword_seqs = []
        for data in test_data:
            content, keword, start, end = data
            context_seq = self.tokenizer.text_to_sequence(content,max_seq_len=self.max_seq_len)
            keword_seq = self.tokenizer.text_to_sequence(keword,max_seq_len=self.max_seq_len)
            context_seqs.append(context_seq)
            keword_seqs.append(keword_seq)
        context_indices = torch.tensor(context_seqs, dtype=torch.int64).to(self.device)
        keword_indices = torch.tensor(keword_seqs, dtype=torch.int64).to(self.device)

        t_inputs = [context_indices, keword_indices]
        t_outputs = self.model(t_inputs)

        t_probs = F.softmax(t_outputs, dim=-1).cpu().detach().numpy()
        label_idx = t_probs.argmax(axis=-1)
        label_num = [self.label_list[idx] for idx in label_idx]

        label_name = [idx == 1 for idx in label_num]
        return label_name


def test_one():
    """
    测试一条数据
    """
    model = LabelComponentModel()
    content = '减少癌症的发生几率，改善癌症的病情。它含有的茶多酚能有效的阻止一些致癌物质的合成，如亚硝酸铵，对癌细胞还有着破坏的作用。'
    keword = '茶多酚'
    start = 22
    end = 25
    label_result = model.predict(content, keword, start, end)
    print(content)
    print(label_result)


def test_batch():
    """
    测试一条数据
    """
    model = LabelComponentModel()
    test_data = [
        ['减少癌症的发生几率，改善癌症的病情。它含有的茶多酚能有效的阻止一些致癌物质的合成，如亚硝酸铵，对癌细胞还有着破坏的作用。', '茶多酚', 22, 25],
        ['也是品牌的产品，购买放心。是秒杀价时购买的。价位合适。是花兰花香型的。家里男生和女生都适用。使用时泡沫丰富，发质丝滑，', '兰花', 22, 25],
        ['减少癌症的发生几率，改善癌症的病情。它含有的茶多酚能有效的阻止一些致癌物质的合成，如亚硝酸铵，对癌细胞还有着破坏的作用。', '茶多酚', 22, 25]
    ]
    label_result = model.predict_batch(test_data)
    for idx, d in enumerate(test_data):
        print("----")
        print(d[0])
        print(label_result[idx])


if __name__ == '__main__':
    test_one()
    # test_batch()


# -*- coding: utf-8 -*-
# ----------------------------------------------
# author : 
# create_time : 2020/6/29 14:10
# update_time : 2020/6/29 14:10
# file : time_format.py
# copyright : Lavector
# ----------------------------------------------
"""
提取时间并格式化
"""
import datetime
import collections

from lavda.backexcutor.analysis_interface import RowInterface


class DateTimeFormat(RowInterface):
    """
    从文本中提取时间数据并格式化
    """

    def __init__(self, format_str="%Y-%m-%d", default_value=None):
        """

        :param format_str: str, 时间格式
        :param default_value: 任何值, 默认值
        """
        super(DateTimeFormat, self).__init__(default_value)

        self.default_value = default_value
        self.format_str = format_str
        # TODO '十'要特殊处理，其他处就要全部删掉
        data_map_list = [('十月', '10月'), ('二十日', "20日"), ('三十日', '30日'), ('十日', '10日'), ('十', ''),
                         ('〇', '0'), ('○', '0'), ('一', '1'), ('二', '2'), ('三', '3'), ('四', '4'),
                         ('五', '5'), ('六', '6'), ('七', '7'), ('八', '8'), ('九', '9'),
                         ('+', ''), ('/', '-'), ('.', '-'), ('年', '-'), ('月', '-'), ('日', '')
                         ]
        self.data_map = collections.OrderedDict(data_map_list)

        self.effective_char = set([' ', ':', '-'] + list('0123456789'))

    def is_all_effective_char(self, datetime_string):
        for ch in list(datetime_string):
            if ch in self.effective_char:
                continue
            return False
        return True

    def parse(self, data, **params):
        """

        :param data: string
        :param params:
        :return: datetime:
        """
        if data is None:
            return self.default_value
        datetime_string = str(data).strip()
        if not datetime_string:
            return self.default_value

        # TODO 包括毫秒的情况，要去掉，不处理秒以下的数据
        if "." in datetime_string:
            t_index = datetime_string.index(".")
            datetime_string = datetime_string[:t_index]

        empty_char_count = datetime_string.count(" ")
        if empty_char_count > 1:
            datetime_string = data.replace(" ", "", empty_char_count - 1)

        # 无效字符替换
        for key in self.data_map:
            if key in datetime_string:
                datetime_string = datetime_string.replace(key, self.data_map[key])

        # TODO 再一次进行校验
        if len(datetime_string) > 20:
            return self.default_value
        if not self.is_all_effective_char(datetime_string):
            return self.default_value

        # TODO 没有日期数据时，会给一个默认的日期 1900-01-01
        # TODO 没有时间数据时，会给一个默认的时间 00:00:00
        result_date = self.default_value
        if len(datetime_string) <= 10:
            if datetime_string.count("-") == 2:
                # 只包含日期
                result_date = datetime.datetime.strptime(datetime_string, u"%Y-%m-%d")
            elif 6 <= len(datetime_string) <= 8:
                result_date = datetime.datetime.strptime(datetime_string, u"%Y%m%d")
        elif len(datetime_string) > 10 and ":" in datetime_string:
            datetime_string = datetime_string.replace(" ", "")
            # 包含 日期 + 时间
            if ("-" in datetime_string):
                if datetime_string.count(":") == 1:
                    # 时间包含年，月
                    result_date = datetime.datetime.strptime(datetime_string, u"%Y-%m-%d%H:%M")
                elif datetime_string.count(":") == 2:
                    # 时间包含年，月，日
                    result_date = datetime.datetime.strptime(datetime_string, u"%Y-%m-%d%H:%M:%S")
            else:
                if datetime_string.count(":") == 1:
                    # 时间包含年，月
                    result_date = datetime.datetime.strptime(datetime_string, u"%Y%m%d%H:%M")
                elif datetime_string.count(":") == 2:
                    # 时间包含年，月，日
                    result_date = datetime.datetime.strptime(datetime_string, u"%Y%m%d%H:%M:%S")

        return result_date

    def process(self, data, **params):
        if not data:
            return self.default_value
        result_date = self.parse(data, **params)
        if not result_date:
            return self.default_value
        result = result_date.strftime(self.format_str)

        return result


if __name__ == '__main__':
    datetime_format = DateTimeFormat()

    # 1.测试提取时间
    # datetime_string = "2019年4月5日 18:21"
    datetime_string = "2019年06月02日22:50"
    # datetime_string = "2019/06/02 22:50:00.000"
    # datetime_string = "二○一七年六月二十日"
    datetime_string = "2019-10-11 22:13:00"
    result1 = datetime_format.process(datetime_string)
    print(result1)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : dada
# @time   : 2020/8/17 14:45
"""
* 表示支持所有的sheetname
如需要增加打标字段直接增加即可
例如："tmall": {
        "品牌": ['content', 'title'，'keyword'],
    },
    天猫在打品牌标签时就会合并content，title，keyword字段进行打标
"""
tag_choice_dic = {
    "redbook": {
        "*": ['content', 'title'],
    },
    "zhihu": {
        "*": ['content', 'title'],
    },
    "xinyang": {
        "*": ['content', 'title'],
    },
    "tmall": {
        "品牌": ['content', 'title'],
    },
    "jd": {
        "品牌": ['content', 'title'],
    },
    "nosetime": {
        "品牌": ['content', 'title'],
    },
}



# -*- coding: utf-8 -*-
# ----------------------------------------------
# author : 
# create_time : 2020/6/29 14:20
# update_time : 2020/6/29 14:20
# file : gender_format.py
# copyright : Lavector
# ----------------------------------------------
from lavda.backexcutor.analysis_interface import RowInterface
from lavda.util.mysql_util import *
import numpy as np
class GenderFormat(RowInterface):
    """
    提取并格式化性别数据

    """
    def __init__(self, data_dict=None,
                 default_value=None):
        """
        :param data_dict: dict, 映射数据
        :param default_value: int, 无数据或非法数据时返回此默认值
        """
        super(GenderFormat, self).__init__(default_value)

        if data_dict is None:
            self.data_dict = {'男': {'1', '1.0', '男'}, '女': {'2', '2.0', '女'}}
        else:
            self.data_dict = data_dict

        self.default_value = default_value
        self.boy_list = ['直男', '宅男', '阳光男', '眼镜男', '型男', '天蝎男', '天枰男', '天平男', '水瓶男', '双子男', '双鱼男',
                         '熟男', '狮子男', '射手男', '山羊男', '瓶子男', '暖男', '男子汉', '男子', '男纸', '男同学', '男生', '男神经', '男神',
                         '男模', '男护士', '男孩', '男闺蜜', '男高中生', '男粉', '男的', '男博主', '牧羊男', '摩羯男', '猛男', '美男', '理工男', '巨蟹男',
                         '金牛男', '肌肉男', '机车男', '工科男', '奋斗男', '草食男', '白羊男']
        self.girl_list = ['直女', '宅女', '眼镜女', '仙女', '天蝎女', '天枰女', '天平女', '水瓶女', '双子女', '双鱼女', '熟女',
                          '狮子女', '射手女', '少女', '山羊女', '瓶子女', '女子', '女纸', '女友粉', '女同学', '女生', '女神经', '女神', '女模', '女护士',
                          '女汉子', '女孩', '女高中生', '女粉', '女的', '女博主', '牧羊女', '摩羯女', '美女', '巨蟹女', '金牛女',
                          '机车女', '白羊女']
        self.boy_other_list = ['boy', '校草', '小哥哥', '帅哥', '少年']
        self.girl_other_list = ['girl', '公主', '富婆', '妈粉', '校花', '小姐姐', '小姑凉']

    def gender_judgment(self, text):
        '''
        后加的通过简介，用户标签等去判断性别
        :param text:
        :return:
        '''

        if '男' in text:
            boy_index = text.index('男')
            for boy_i in self.boy_list:
                if boy_i in text[boy_index - 2 if boy_index - 2 >= 0 else 0:boy_index + 4]:
                    return '男'
        elif '女' in text:
            girl_index = text.index('女')
            for girl_i in self.girl_list:
                if girl_i in text[girl_index - 2 if girl_index - 2 >= 0 else 0:girl_index + 4]:
                    return '女'
        else:
            for girl_i in self.girl_other_list:
                if girl_i in text:
                    return '女'
            for boy_i in self.boy_other_list:
                if boy_i in text:
                    return '男'
        return

    def process(self, data, params=None):
        column_name_list = ['gender', 'introduce', 'user_label', 'nickname']
        for column_name in column_name_list:
            if column_name not in data:
                data[column_name] = ''
        data_gender = data['gender']

        data_str = str(data_gender).strip()
        result_value = None
        for key in self.data_dict:
            if data_str in self.data_dict[key]:
                result_value = key
                break

        # TODO 2021.8.13 增加了这里的性别判断
        if result_value is None:
            gender_judgment_text = ''.join([str(data[i]) for i in ['introduce', 'user_label', 'nickname'] if data[i]])
            result_value = self.gender_judgment(gender_judgment_text)
            result_value = self.default_value if result_value is None else result_value
        return result_value


if __name__ == '__main__':
    gender_format = GenderFormat()
    print(gender_format.process({'gender':'','nickname':'宅男'}))



# -*- coding: utf-8 -*-
# ----------------------------------------------
# author : 
# create_time : 2020/6/29 15:27
# update_time : 2020/6/29 15:27
# copyright : Lavector
# ----------------------------------------------
from lavda.util.lang.langconv import Converter
from lavda.backexcutor.analysis_interface import *


class LangTool(object):
    @staticmethod
    def simple2tradition(text):
        # 将简体转换成繁体
        result = Converter('zh-hant').convert(text)
        return result

    @staticmethod
    def tradition2simple(text):
        # 将繁体转换成简体
        result = Converter('zh-hans').convert(text)
        return result


class Simple2Tradition(RowInterface):
    """
    简体转繁体
    """
    def __init__(self):
        super(Simple2Tradition, self).__init__(None)

    def process(self, data, **params):
        if data is None:
            return None

        if isinstance(data, dict):
            result = dict()
            for key in data:
                result[key] = Converter('zh-hant').convert(str(data[key]))
        else:
            result = Converter('zh-hant').convert(str(data))

        return result


class Tradition2Simple(RowInterface):
    """
    繁体转简体
    """
    def __init__(self):
        super(Tradition2Simple, self).__init__(None)

    def process(self, data, **params):
        if data is None:
            return None

        if isinstance(data, dict):
            result = dict()
            for key in data:
                result[key] = Converter('zh-hans').convert(str(data[key]))
        else:
            result = Converter('zh-hans').convert(str(data))

        return result


if __name__ == '__main__':
    text = "中国好"
    s = Simple2Tradition()
    result1 = s.process(text)
    print(result1)

    t = Tradition2Simple()
    result2 = t.process(result1)
    print(result2)


# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose :
# author :
# create_time : 2020/7/2 17:21
# update_time : 2020/7/2 17:21
# copyright : Lavector
# ----------------------------------------------
#!/usr/bin/python
# coding:utf-8
from __future__ import print_function
from setuptools import setup, find_packages
import sys
import datetime
import os


version_path = os.path.join(os.getcwd(), *['maxda', "version.py"])
if os.path.exists(version_path):
    version_file = open(version_path)
    cur_version = version_file.readline().strip()
else:
    cur_version = "0.0.5"

setup(
    name="lavda",
    version=cur_version,
    author="lavector data_analysis",
    author_email="hao.jia@lavector.com",
    description="data process and analysis",
    # license="MIT",
    # url="http://gitlab.lavector.com/analysis/maxda-dev.git",
    packages=find_packages(exclude=["configs"]),
    classifiers=[
        "Environment :: Web Environment",
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    install_requires=[
        'pandas',
        'numpy',
        'xlsxwriter',
        'joblib>=0.14.0',
        'xlrd',
        'pymysql',
        'emoji',
        "sklearn",
        "pyahocorasick",
        "openpyxl",
        "torch>=1.2.0"
    ],
    zip_safe=True,
)





# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose : 
# author : 
# create_time : 2020/7/1 16:19
# update_time : 2020/7/1 16:19
# copyright : Lavector
# ----------------------------------------------
import pandas as pd
import collections

from lavda.analysis.da_tag.tag_dict import PandaLabelDictTool


def format_dict2list(r_dict):

    l_tup = sorted(r_dict.items(), key=lambda item: item[0], reverse=True)
    r_list = []
    max_num = 0
    for k, v in l_tup:
        if v["level"] > max_num:
            max_num = v["level"]
        result = []
        if v["level"] == max_num:
            result.append(v["text"])
            recursion(r_dict, result, v["parent"])
            r_list.append(result[::-1])
    return r_list

def recursion(r_dict, result, id):
    ps = r_dict[id]
    text, parent  = ps["text"], ps["parent"]
    result.append(text)
    if parent in r_dict:
        recursion(r_dict, result, parent)


def au_get_tree2dict(filename,sheet_name,level):
    """
    返回数据库标准存储结构
    """
    dict_tool = PandaLabelDictTool()
    label_dict_object = dict_tool.load_tree_dict(
        filename,
        sheet_name=sheet_name,
        header=False,
        tree_format=0,
        total_label_level=level,
        label_name='beauty',
        split_text=False)
    r_dict = label_dict_object.get_dict()
    return r_dict
def format_list2df(result):
    max_len = 0
    for i in result:
        if len(i) > max_len:
            max_len = len(i)


    data = []
    for i in result:
        loss = max_len - len(i)
        tmp_data = i[:-1] + ["" for _ in range(loss)] + [i[-1].split("|")[0], i[-1]]
        data.append(tmp_data)

    max_len += 1
    df = pd.DataFrame(data)
    columns = ["level_%s" % i for i in range(1, max_len+1)]
    columns[-2] = "keytag"
    columns[-1] = "keyword"
    df.columns = columns
    return df


def au_get_tree(infile, sheet,level,sheet_name):
    r_dict = au_get_tree2dict(infile,sheet,level)
    r_list = format_dict2list(r_dict)
    df = format_list2df(r_list)
    df = df.drop(["keytag"], axis=1)
    df["keyword"] = df["keyword"].apply(lambda line: str(line).replace(" ", "").lower())
    df = df.sort_values(by=list(df.columns), ascending=True)
    columns_dict = au_get_new_columns(level,sheet_name)
    df.rename(columns=columns_dict, inplace=True)
    tree_list = [tree for line,tree in columns_dict.items() ]
    return df,tree_list,sheet_name



def au_get_line(path,sheet):
    df = pd.read_excel(path, sheet, error_bad_lines=False, encoding="utf_8_sig")
    line_new_list = []
    line_list = list(df.columns)
    columns_dict = {}
    the_pop1 = 'keyword'    # 后续代码中用的都是keyword，所以我这儿就不改了
    the_pop2 = [i for i in line_list if 'keyword' in i.replace(" ", "").lower()][0]     # 这儿是给的词表里keyword这列的字段名
    for line in line_list:
        new_line = "{}({})".format(sheet, line)
        columns_dict[line] = new_line
        line_new_list.append(new_line)
    new_keyword1 = "{}({})".format(sheet, the_pop1)
    new_keyword2 = "{}({})".format(sheet, the_pop2)
    columns_dict[the_pop1] = new_keyword1
    line_new_list.append(new_keyword1)
    line_new_list.remove(new_keyword2)
    df.rename(columns=columns_dict, inplace=True)
    df[new_keyword1] = df[new_keyword2].apply(lambda line: str(line).replace(" ", "").lower())
    if the_pop2 != the_pop1:
        df.drop(new_keyword2, axis=1, inplace=True)
    return df, line_new_list, sheet




def au_get_new_columns(level,sheet_name):
    columns_dict = collections.OrderedDict()
    level -= 1
    for col in range(level):
        str_col = str(col+1)
        columns_dict['level_{}'.format(str_col)] = '{}(level{})'.format(sheet_name,str_col)
    columns_dict['keyword'] = "{}(keyword)".format(sheet_name)
    return columns_dict


# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose : 核心配置功能
# author : 
# create_time : 2020/7/2 13:37
# update_time : 2020/7/2 13:37
# copyright : Lavector
# ----------------------------------------------
import os

# 数据分析数据库
CONF_DB_LAVWIKI = {'host': "111.207.114.228",
                   'port': 13306,
                   'user': "data_analysis",
                   'password': "data_analysis_root_0715",
                   'database': "lavwiki",
                   'charset': "utf8",
                  }

#分句符号配置

split_sign = '[.。!！\n]'

# AI 数据库
LAV_AI_DB_CONFIG = {"host": "192.168.50.129",
                    "port": 3306,
                    "user": "lavdata2019",
                    "password": "lav_data_20190905",
                    "database": "lav_ai",
                    }

# NLP 模型配置
def get_nlp_model_path():
    """
    获取nlp 模型根目录
    :return:
    """
    can_model_paths = ['/data/share/nlp', 'C:\\jh\\share\\nlp', 'E:\\data\\share\\nlp']

    for model_path in can_model_paths:
        if os.path.exists(model_path):
            return model_path

    raise Exception("can not find nlp model path")



# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose : 清洗接口
# author : 
# create_time : 2020/7/1 15:39
# update_time : 2020/7/1 15:39
# copyright : Lavector
# ----------------------------------------------
from lavda.backexcutor.analysis_interface import *


class CleanInterface(object):
    pass


class CleanFlag:
    """
    清洗状态标志
    """
    YES = 1
    NO = 0



# -*- coding: utf-8 -*-
# ----------------------------------------------
# author : 
# create_time : 2020/6/29 14:24
# update_time : 2020/6/29 14:24
# copyright : Lavector
# ----------------------------------------------
from lavda.backexcutor.analysis_interface import RowInterface


class IntFormat(RowInterface):
    """"
    整数格式化
    """

    def __init__(self, default_value=""):
        """
        :param default_value: 默认值
        """
        super(IntFormat, self).__init__(default_value)
        self.default_value = default_value

    def process(self, data, **params):
        if data is None:
            return self.default_value
        if isinstance(data, int):
            return data

        data = str(data).strip()
        if not data:
            return self.default_value

        if '千' in data:
            data = data.replace('千', '')
            data = int(float(data) * 1000)
            return data
        elif '万' in data:
            data = data.replace('万', '')
            data = int(float(data) * 10000)
            return data
        elif '亿' in data:
            data = data.replace('亿', '')
            data = int(float(data) * 100000000)
            return data

        # 判断下是不是数字,是数字就保留
        if u'\u0030' <= data <= u'\u0039':
            return int(float(data))
        else:
            return self.default_value


class FloatFormat(RowInterface):
    """
    浮点数格式化
    """
    def __init__(self, n_digits=2, default_value=""):
        """

        :param n_digits: int, 小数点后保留位数
        :param default_value: 默认值
        """
        super(FloatFormat, self).__init__(default_value)

        self.n_digits = n_digits
        self.default_value = default_value

    def process(self, data, **params):
        """
        1：将各个字段的数据进行单位格式化，例如将'万'，'亿'等单位替换成纯数字。
        2：将不是数字的替换成空。
        3: 将price字段处理成保留两位的浮点数。
        :param data: 单个字段的值
        :return: 处理后单个字段的值
        """
        if data is None:
            return self.default_value
        data = str(data).strip()
        if not data:
            return self.default_value

        if '千' in data:
            data = data.replace('千', '')
            data = float(data) * 1000
            return data
        elif '万' in data:
            data = data.replace('万', '')
            data = float(data) * 10000
            return data
        elif '亿' in data:
            data = data.replace('亿', '')
            data = float(data) * 100000000
            return data

        # 判断下是不是数字,是数字就保留
        if u'\u0030' <= data <= u'\u0039':
            return round(float(data), self.n_digits)
        else:
            return self.default_value


if __name__ == '__main__':
    int_format = IntFormat()
    print(int_format.process("23.23"))

    float_format = FloatFormat()
    print(float_format.process("23.2390"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from lavda.analysis.da_clean.clean_interface import *
from lavda.backexcutor.analysis_interface import RowInterface
from lavda.util.panda_tool import PandaTool


class CleanUnincludedKeyword(RowInterface, CleanFlag):
    
    def __init__(self):
        super(CleanUnincludedKeyword, self).__init__(CleanFlag.NO)

        self.keyword_list = None

    def load_config_data(self, **config):

        keyword_path = config['KeywordExcelPath']
        # 读取配置文件中抓取关键词
        keyword_df = PandaTool.read(keyword_path)
        keyword_list = keyword_df.iloc[:, 0].values.tolist()
        self.keyword_list = keyword_list

    def process(self, data, **params):

        for keyword_i in self.keyword_list:
            if str(keyword_i).strip().lower() in str(data).lower():
                return CleanFlag.NO
        return CleanFlag.YES



# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose : 
# author : 
# create_time : 2020/6/30 18:27
# update_time : 2020/6/30 18:27
# copyright : Lavector
# ----------------------------------------------
import importlib
from lavda.util.file_tool import FileTool


class PyModelLoader(object):
    def __init__(self):
        package_list = list()
        package_list.append('lavda.analysis.da_pre')
        # package_list.append('maxda.analysis.da_save')
        package_list.append('lavda.analysis.da_extract')
        package_list.append('lavda.analysis.da_clean')
        package_list.append('lavda.analysis.da_tag')
        package_list.append('lavda.analysis.da_statis')

        m_list = list()
        for p_name in package_list:
            p_m = importlib.import_module(p_name)
            # TODO 每次使用时都重新加载，这样maxda更新时，不再需要更新autoda
            # importlib.reload(p_m)

            file_name_list = FileTool.get_all_file(p_m.__path__[0], suffix='.py')
            file_name_list = [FileTool.get_base_name(name) for name in file_name_list]
            file_name_list = [name.replace(".py", "") for name in file_name_list]
            file_name_list.remove("__init__")

            for py_name in file_name_list:
                py_m = importlib.import_module(p_name + '.' + py_name)
                m_list.append(py_m)

        self.m_list = m_list

    def load(self, class_name, **init_param):
        for m_py in self.m_list:
            if not hasattr(m_py, class_name):
                continue
            obj_class_name = getattr(m_py, class_name)
            if init_param is None:
                target_class = obj_class_name()
            else:
                target_class = obj_class_name(**init_param)
            return target_class

        return None



# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose: 广告垃圾检测
# author : 
# create_time : 2020/6/29 15:32
# update_time : 2020/6/29 15:32
# copyright : Lavector
# ----------------------------------------------
"""
广告清洗
"""
import re
import codecs
import numpy as np

from lavda.util.mysql_util import *
from lavda.analysis.da_clean.clean_interface import *


class SpamInterface(object):
    def predict(self, text, params=None):
        """
        :param text:
        :param params:
        :return: bool
        """
        pass


class OrdinaryWays(object):
    @staticmethod
    def get_from2mysql(c_type="adword"):
        sql = """
        select keyword from %s
        """ % c_type
        result = np.array(LavwikiReader.execute_all(sql)[0])
        result = [i[0] for i in result]
        return result

    @staticmethod
    def get_clean(c_type):
        """
        type:mysql数据库表名，处理weibo垃圾、水军时的词

        """
        clean = OrdinaryWays.get_from2mysql(c_type)
        clean = [str(i).lower().strip() for i in clean if len(str(i).strip()) > 0]
        return clean

    @staticmethod
    def read_by_line(file_path):
        """
        按行读取数据
        :param file_path:
        :return:
        """
        data_list = list()
        with codecs.open(file_path, 'r', encoding="utf-8") as fp:
            for line in fp:
                line = line.strip()
                if len(line) > 0:
                    data_list.append(line)

        return data_list

    @staticmethod
    def has_chinese_count(text):
        """
        判断中文字符数量
        :param self:
        :param text:
        :return:
        """
        count = 0
        tmp_text = text
        if isinstance(tmp_text, str):
            tmp_text = str(tmp_text)
        for ch in tmp_text:
            if '\u4e00' <= ch <= '\u9fff':
                count += 1
        return count

    @staticmethod
    def is_num(text):
        """
        判断是否为数字，包括整数和浮点数
        :param text:
        :return:
        """
        # return text.isdigit()

        # pattern = re.compile(r'^[-+]?[-0-9]\d*\.\d*|[-+]?\.?[0-9]\d*$')
        # result = pattern.match(text)
        # if result:
        #     return True
        # else:
        #     return False

        if text.replace('.', '', 1).isdigit():
            return True
        return False

    @staticmethod
    def get_spam_dict(dict_file_path):
        word_set = set()
        # 1.加载广告词
        word_list = OrdinaryWays.read_by_line(dict_file_path)
        for word in word_list:
            tmp_word = word.strip()
            if tmp_word:
                # TODO 全部转小写
                word_set.add(tmp_word.lower())

        return word_set


class SpamPatternRule(SpamInterface):
    """
    基于正则规则判断，检查是否包含特定的正则表达式
    """
    def __init__(self):
        self.ads_item_rule_list = list()
        self._load()

    def _load(self):
        tmp_rule_set = OrdinaryWays.get_clean('weibo_ads_pattern_rule')
        for tmp_rule in tmp_rule_set:
            tmp_rule_pattern = re.compile(tmp_rule, re.S)
            self.ads_item_rule_list.append(tmp_rule_pattern)

    def predict(self, text, params=None):
        for re_rule_pattern in self.ads_item_rule_list:
            result = re_rule_pattern.search(text)
            if result and result.group() is not None:
                return True
        return False


class SpamHastagRule(SpamInterface):
    """
    基于 Hasttag 中的内容进行判断
    处理类型同pattern rule
    """

    def __init__(self, other_rule_list=None):
        """
        :param dict_file_path:
        :param other_rule_list: 其他自定义的规则
        """
        self.item_rule_list = list()
        self.other_rule_list = other_rule_list

        self._load()

    def _load(self):
        tmp_rule_set = OrdinaryWays.get_clean('weibo_ads_hastag_rule')

        for tmp_rule in tmp_rule_set:
            tmp_rule_pattern = re.compile(tmp_rule, re.S)
            self.item_rule_list.append(tmp_rule_pattern)

    def is_match(self, text, params=None):
        for re_rule_pattern in self.item_rule_list:
            result = re_rule_pattern.search(text)
            if result and result.group() is not None:
                return True
        return False

    def predict(self, text, params=None):
        r2_list = self._get_bracket(text, (u'【', u'】'))
        bracket_list = r2_list

        for item in bracket_list:
            tmp_item = item.strip()
            if self.other_rule_list:
                for other_rule in self.other_rule_list:
                    if other_rule.predict(tmp_item):
                        return True

            if self.is_match(tmp_item):
                return True
        return False

    def _get_bracket(self, text, bracket, drop_bracket=True):
        tmp_text = text
        re_string = u"[{}].*?[{}]".format(bracket[0], bracket[1])
        p1 = re.compile(re_string, re.S)  # 最小匹配
        find_list = re.findall(p1, tmp_text)

        result_list = list()
        if find_list:
            for item in find_list:
                tmp_item = item
                if drop_bracket:
                    tmp_item = tmp_item.replace(bracket[0], u'')
                    tmp_item = tmp_item.replace(bracket[1], u'')
                # TODO,由于抓取数据解析问题，可能许多符号会显示为？号，因此全替换掉
                tmp_item = tmp_item.replace(u'?', u'')
                result_list.append(tmp_item)
        return result_list


class ItemPriceRule(SpamInterface):
    """
    判断是否表示价格
    """
    def __init__(self):
        price_rule_str_list = [u'(￥|\$)\d+(\.\d+)?',
                               u'\d+(\.\d+)?(元|入)']
        rule_str = u'|'.join(price_rule_str_list)
        rule_str = u'(' + rule_str + u')'

        self.price_rule_pattern = re.compile(rule_str, re.S)

    def predict(self, text, params=None):
        """
        判断是否表示价格,只是表示可能是价格
        :param text:
        :return:
        """
        tmp_text = text
        # 1.根据正则查找
        tmp_text = tmp_text.replace(u' ', u'')

        result = self.price_rule_pattern.search(tmp_text)
        if result and result.group() is not None:
            return True

        # 2.是否为【】包起来的数字
        if u'【' in tmp_text or u'】' in tmp_text:
            tmp_item = tmp_text.replace(u'【', u'')
            tmp_item = tmp_item.replace(u'】', u'')
            if OrdinaryWays.is_num(tmp_item):
                return True

        return False

class ItemNumRule(SpamInterface):
    """
    判断是否为数字
    """
    def predict(self, text, params=None):
        if OrdinaryWays.is_num(text):
            return True

        return False


class SpamWordRule(SpamInterface):
    """
    根据是否包含特定的词来判断
    """

    def __init__(self,use_re=False):
        """

        :param dict_file_path:
        :param use_re: 是否使用正则表达式
        """
        self.word_set = set()
        self.rule_pattern = None

        self._load()

        self.use_re = use_re

    def _load(self):
        self.word_set = OrdinaryWays.get_clean('weibo_ads_word')
        self.rule_pattern = re.compile(u'|'.join(self.word_set), re.S)

    def _do_by_re(self, text, params=None):
        result = self.rule_pattern.search(text)
        if result and result.group() is not None:
            return True

        return False

    def _do(self, text, params=None):
        for word in self.word_set:
            if word in text:
                return True

        return False

    def predict(self, text, params=None):
        if self.use_re:
            return self._do_by_re(text)
        return self._do(text)


class AdsDelimiter(SpamInterface):
    """
    按照分隔符，来判断是否为广告
    """
    # 注意 最后一个是空格符
    SPLIT_CHAR = u'， '
    MAX_WORD_LEN = 30

    def __init__(self, rule_price):
        self.rule_price = rule_price

        self.item_rule_list = list()

        rule_str_set = OrdinaryWays.get_clean("weibo_ads_delimiter_rule")
        for tmp_rule in rule_str_set:
            tmp_rule_pattern = re.compile(tmp_rule, re.S)
            self.item_rule_list.append(tmp_rule_pattern)

    def match_by_pattern(self, text, params=None):
        """
        根据正则匹配查找
        :param text:
        :param params:
        :return:
        """
        for re_rule_pattern in self.item_rule_list:
            result = re_rule_pattern.search(text)
            if result and result.group() is not None:
                return True
        return False

    def predict(self, text, params=None):
        chinese_char_count = OrdinaryWays.has_chinese_count(text)
        if chinese_char_count > self.MAX_WORD_LEN:
            return False

        # 拆分数据，进行处理
        re_string = u'[{}]'.format(self.SPLIT_CHAR)
        text_list = re.split(re_string, text)
        text_list = [item.strip() for item in text_list if item.strip()]

        if len(text_list) < 2:
            return False

        # 根据广告词判断
        if self.match_by_pattern(text):
            return True

        # 根据是否有价格标识
        for item in text_list:
            # 几乎一定是价格
            if self.rule_price.predict(item) and len(item) < 6:
                return True
            # 是数字，但可能是价格
            if OrdinaryWays.is_num(item.strip()) and len(text_list) < 6:
                return True
            # 英文 off,eg:Burberry vip 50% off 只有355/36 ?
            if u'off' in item:
                return True

        return False


class AdsLineRule(SpamInterface):
    """
    以行为单位判断是否为广告
    """

    def __init__(self, rule_list):
        self.rule_list = rule_list

    def predict(self, text, params=None):
        sub_lines = text.split(u'\n')
        sub_lines = [item.strip() for item in sub_lines if item.strip()]
        for line in sub_lines:
            for rule in self.rule_list:
                if rule.predict(line):
                    return True
        return False


class AdsSpecialRule(SpamInterface):
    """
    特别规则，只针对几个特殊字符
    放一些不成熟的规则，后期有很大概率要调整的
    """
    SPECIAL_CHAR = [u'💰', u'💵', u'$']

    def predict(self, text, params=None):
        for item in self.SPECIAL_CHAR:
            if item in text:
                return True

        # 格式不全
        if u'【' in text and u'】' not in text:
            return True
        if u'【' not in text and u'】' in text:
            return True

        return False


class AdsPrefixRule(SpamInterface):
    """
    基于文本前缀判断
    """

    def __init__(self):
        self.base_rule = SpamPrefixRule()

    def predict(self, text, params=None):
        lines = params
        if len(lines) < 0:
            return False
        # 第一行
        if self.base_rule.predict(lines[0].strip()):
            return True

        # 第二行
        if self.base_rule.predict(lines[-1].strip()):
            return True

        return False


class SpamPrefixRule(SpamInterface):
    """
    基于文本前缀判断
    """
    def __init__(self, user_re=False):
        self.word_set = set()
        self.rule_pattern = None

        self._load()

        self.use_re = user_re

    def _load(self):
        # 1.加载广告词
        self.word_set = OrdinaryWays.get_clean('weibo_ads_prefix_word')

        rule_str = u"^" + u"(" + u'|'.join(self.word_set) + u')'
        self.rule_pattern = re.compile(rule_str, re.I | re.S)

    def _do_by_re(self, text, params=None):
        result = self.rule_pattern.search(text)
        if result and result.group() is not None:
            return True

        return False

    def _do(self, text, params=None):
        for word in self.word_set:
            if text.startswith(word):
                return True

        return False

    def predict(self, text, params=None):
        if self.use_re:
            return self._do_by_re(text)
        return self._do(text)


class AdsSuffixRule(SpamInterface):
    """
    基于文本后缀判断
    """

    def __init__(self):
        self.base_rule = SpamSuffixRule()

    def predict(self, text, params=None):
        """
        该文本是否为客观描述类
        :param text:
        :param params:
        :return:
        """
        lines = params
        if len(lines) < 0:
            return False

        # 第一行
        if self.base_rule.predict(lines[0]):
            return True

        # 最后一行
        if self.base_rule.predict(lines[-1]):
            return True

        return False


class SpamSuffixRule(SpamInterface):
    """
    基于文本前缀判断
    """

    def __init__(self, use_re=False):
        self.word_set = set()
        self.rule_pattern = None

        self._load()

        self.use_re = use_re

    def _load(self):
        self.word_set = OrdinaryWays.get_clean('weibo_ads_suffix_word')

        rule_str = u"(" + u'|'.join(self.word_set) + u')' + u"$"
        self.rule_pattern = re.compile(rule_str, re.I | re.S)

    def _do_by_re(self, text, params=None):
        result = self.rule_pattern.search(text)
        if result and result.group() is not None:
            return True

        return False

    def _do(self, text, params=None):
        for word in self.word_set:
            if text.endswith(word):
                return True

        return False

    def predict(self, text, params=None):
        if self.use_re:
            return self._do_by_re(text)
        return self._do(text)


class WeiboAdsRuleModel(RowInterface, CleanInterface):
    def __init__(self, **params):
        super(WeiboAdsRuleModel, self).__init__(CleanFlag.NO)

        self.rule_pattern = SpamPatternRule()
        self.rule_hastag = SpamHastagRule([ItemPriceRule(), ItemNumRule()])
        self.rule_word = SpamWordRule()
        self.rule_delimiter = AdsDelimiter(ItemPriceRule())

        sub_rule_list = list()
        sub_rule_list.append(self.rule_word)
        sub_rule_list.append(self.rule_pattern)
        sub_rule_list.append(self.rule_delimiter)
        sub_rule_list.append(self.rule_hastag)

        self.rule_list = list()
        self.rule_list.append(AdsLineRule(sub_rule_list))
        self.rule_list.append(AdsSpecialRule())
        self.rule_list.append(AdsPrefixRule())
        self.rule_list.append(AdsSuffixRule())

    def _clean(self, text):
        tmp_text = text
        # 去掉特殊的结尾符(html格式中包含)
        ch = chr(8203)
        tmp_text = tmp_text.replace(ch, u'')
        tmp_text = tmp_text.replace(u'\r\n', u'\n')

        # TODO 去掉多余的问号,有好些问号是解析或抓取时产生的，不是原文自带的
        tmp_text = tmp_text.replace(u'?', u'')

        return tmp_text

    def process(self, data, **params):
        """
        预测
        :param data: 文本
        :param params: 参数
        :return:
        """
        clean_text = str(data).lower()
        lines = clean_text.split(u'\n')
        lines = [line.strip() for line in lines]
        for rule in self.rule_list:
            if rule.predict(clean_text, lines):
                return CleanFlag.YES

        return CleanFlag.NO


def run_spam():
    sub_model_ads = WeiboAdsRuleModel()

    content = u'【19.9】宝丽康美 核桃黑芝麻黑豆代餐粉'
    # content = u'桃黑芝麻黑豆代餐粉'
    channel = u"weibo"
    result = sub_model_ads.process(content)
    print(result)


if __name__ == '__main__':
    run_spam()





# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose : 打标签功能入口
# author : LGD
# create_time : 2021/2/22 09:54
# update_time : 2021/2/22 09:54
# copyright : Lavector
# ----------------------------------------------
import os

from lavda.config.maxda_config import get_nlp_model_path
from lavda.backexcutor.analysis_interface import DfProcessInterface, RowInterface
from lavda.analysis.da_tag.label_extract import KeywordBaseExtractor, KeywordSegmentExtractor, KeywordExtractUtil
from lavda.analysis.da_tag.label_dict import PandaLabelDictTool, LineLabelDict
from lavda.analysis.da_tag.label_verify import VerifyLabelDict
from lavda.analysis.da_tag.label_core import LabelComponentChecker


class LabelBase(RowInterface):
    def __init__(self, extractor, use_model=False,
                 model_root_path=None,
                 default_value=None):
        super(LabelBase, self).__init__((default_value, default_value, default_value))
        self.extractor = extractor
        self.use_model = eval(use_model) if isinstance(use_model, str) else use_model
        if self.use_model:

            if model_root_path is None:
                model_root_path = get_nlp_model_path()

            self.com_checker = LabelComponentChecker(root_model_path=model_root_path)

            self.com_mapping_kid = dict()
        else:
            self.com_checker = None
            self.com_mapping_kid = dict()

    @property
    def mc(self):
        return self.extractor.mc

    def load_config_data(self, **config):
        # todo 计划取消打标前词表验证，改为上传时词表验证
        # tag_config_path = configs['mappingDictPath']
        # ver = VerifyLabelDict(tag_config_path)
        # analyze_result = ver.analyze()
        # if len(analyze_result) > 0:
        #     raise Exception("mapping dict has error")

        self.extractor.load_config_data(**config)

        # 确定模型需要处理的keyword
        # mapping 索引 => kid
        if self.use_model:
            self.com_mapping_kid = dict()
            for i, cur_mapping in enumerate(self.extractor.mc.mappings):
                if isinstance(self.mc, LineLabelDict):
                    # TODO 成分之类不可能在标准词表中出现
                    continue

                id2relation = cur_mapping.id2relation
                sel_kid_set = set()
                for one_kid in id2relation:
                    # 最后两位分别为keyword，和kid，因此要排除掉
                    cur_relation = id2relation[one_kid][:-2]
                    if '成分' in cur_relation:
                        sel_kid_set.add(one_kid)
                if len(sel_kid_set) <= 0:
                    continue
                self.com_mapping_kid[i] = sel_kid_set

    def analyze_semantics(self, input_data_dict, keyword_result):
        """
        语义级标签识别
        :param input_data_dict:
        :param keyword_result:
        :return:
        """
        # 2.语义级判断
        for i, one_result in enumerate(keyword_result):
            # cur_mapping = self.mc.mappings[i]
            if i not in self.com_mapping_kid:
                continue

            # mapping_name = cur_mapping.label_name
            # id2relation = cur_mapping.id2relation
            for col in input_data_dict:
                select_one_result = list()
                non_select_one_result = list()
                for one_item in one_result[col]:
                    # TODO 判断是否为成分

                    if one_item['kid'] in self.com_mapping_kid[i]:
                        select_one_result.append(one_item)
                    else:
                        non_select_one_result.append(one_item)

                last_col_result = non_select_one_result
                if len(select_one_result) > 0:
                    col_content = input_data_dict[col]
                    predict_result = self.com_checker.predict(col_content, select_one_result)
                    predict_one_result = list()
                    for k, k_result in enumerate(predict_result):
                        if k_result:
                            predict_one_result.append(select_one_result[k])

                    last_col_result.extend(predict_one_result)

                one_result[col] = last_col_result

        return keyword_result

    def do_label(self, data, **params):
        # 1.初步提取
        input_data_dict, keyword_result = self.extractor.process(data, **{"merge_and_format": False,
                                                                          "return_all": True})

        # 2.语义级标签识别
        if self.use_model:
            keyword_result = self.analyze_semantics(input_data_dict, keyword_result)

        # 3.结果合并并格式化
        merge_data = self.extractor.merge_and_format(keyword_result, input_data_dict=input_data_dict)
        last_result = self.extractor.build_result_format(keyword_result, merge_data)

        return last_result

    def process(self, data, **params):

        return self.do_label(data, **params)


class LabelDealSimple(LabelBase):
    def __init__(self, use_model=False,
                 model_root_path=None,
                 default_value=None):
        extractor = KeywordBaseExtractor()
        super(LabelDealSimple, self).__init__(extractor, use_model, model_root_path, default_value=default_value)


class LabelDealRelation(LabelBase):
    def __init__(self, use_model=False,
                 model_root_path=None,
                 default_value=None):
        extractor = KeywordSegmentExtractor()
        super(LabelDealRelation, self).__init__(extractor, use_model, model_root_path, default_value=default_value)


class LabelResultFormat(DfProcessInterface):
    def __init__(self, intermediate='tag_result', add_head_text='tag_format_'):
        """

        :param intermediate: string,打标签中间结果字段名
        :param add_head_text: string，格式化后结果字段名前缀
        """
        self.intermediate = intermediate
        self.add_head_text = add_head_text
        self._mc = None

    def load_config_data(self, **config):
        dict_tool = PandaLabelDictTool()

        label_dict_path = config['mappingDictPath']
        self._mc = dict_tool.load_all_from_file(label_dict_path)

    def process(self, data, **params):
        df_format = KeywordExtractUtil.expand_and_format(data, self._mc, tag_result_col=self.intermediate,
                                                         result_prefix=self.add_head_text)


        return df_format









# -*- coding:utf-8 -*-
import findspark

import os
from pyspark import SparkContext
from pyspark import SparkConf
from pyspark.sql import HiveContext, SparkSession
import pandas as pd
# from middleware.utils import get_dir_files

input_path = "/home/wac/hanaojian/maxda-dev/data"
project_id = 'kh_2020-02-30'

global path_root, path_project
path_root = os.getcwd()
path_project = os.path.join(path_root, "data/run_project")
input_data_path = os.path.join(path_project, "1_log")


def get_dir_files(dir, filelist=[], status=0, str1=""):
    """
    dir: 目录地址
    filelist: 变量地址
    status: 表示是包含 1，取消 0，还是不包含 -1
    str1: 过滤词
    """
    newDir = dir
    if os.path.isfile(dir):
        filelist.append(dir)
    elif os.path.isdir(dir):
        for s in os.listdir(dir):
            # 如果需要忽略某些文件夹，使用以下代码
            if status == 1:
                # 如果不包含改字符串就跳过
                if str1 not in s:
                    continue
            elif status == -1:
                # 如果包含改字符串就跳过
                if str1 in s:
                    continue
            newDir = os.path.join(dir, s)
            get_dir_files(newDir, filelist, status, str1)
    # 默认顺序排列
    return sorted(filelist, reverse=False)


def mergefile(input_path_1, project_id):
    files_path = get_dir_files(input_path_1, [], status=-1, str1=".DS_Store")
    print(files_path)
    df_list = []
    for infile in files_path:
        df = pd.read_excel(infile, header=0, error_bad_lines=False, encoding="utf_8_sig",dtype='str').fillna("")
        df_list.append(df)
    data_pd = pd.concat(df_list)
    data_pd['project_id'] = project_id

    return data_pd


def main():
    findspark.init()
    conf = SparkConf().setAppName('savepyspark').setMaster('local[*]')
    sc = SparkContext(conf = conf)
    sqlContext = HiveContext(sc)

    data_pd = mergefile(input_path,project_id)
    df_1 = sqlContext.createDataFrame(data_pd)

    df_1.write.saveAsTable('test.write_test10', mode='append', partitionBy=['project_id'])

    sc.stop()


if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose : 
# author : 
# create_time : 2020/7/1 16:41
# update_time : 2020/7/1 16:41
# copyright : Lavector
# ----------------------------------------------
import re

def remove_emoji(text):
    emoji_pattern = re.compile("["
                               u"\U0001F600-\U0001F64F"  # emoticons
                               u"\U0001F300-\U0001F5FF"  # symbols & pictographs
                               u"\U0001F680-\U0001F6FF"  # transport & map symbols
                               u"\U0001F1E0-\U0001F1FF"  # flags (iOS)
                               "]+", flags=re.UNICODE)

    return emoji_pattern.sub(r'', text)


def DBC2SBC(ustring):
    """
    全角转半角
    https://www.jb51.net/article/124269.htm
    :param ustring:
    :return:
    """
    rstring = ""
    for uchar in ustring:
        inside_code = ord(uchar)
        if inside_code == 0x3000:
            inside_code = 0x0020
        else:
            inside_code -= 0xfee0
        if not (0x0021 <= inside_code and inside_code <= 0x7e):
            rstring += uchar
            continue
        rstring += chr(inside_code)
    return rstring


# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose : 
# author : LGD
# create_time : 2021/4/12 14:14
# update_time : 2021/4/12 14:14
# copyright : Lavector
# ----------------------------------------------


class DictTrieNode(object):
    # 中间节点
    STATUS_MIDDLE = 1
    # 结束节点(实际上也可能是树的中间节点)
    STATUS_END = 2

    def __init__(self, node_value=None, parent_node=None):
        # 节点值
        self.node_value = node_value
        # 父节点：DictTrieNode
        self.parent_node = parent_node
        # 子节点
        self.child_node_dict = dict()
        # 节点元数据
        self.meta_data = None
        # 节点状态
        self.node_status = self.STATUS_MIDDLE

    def get_parent_node(self):
        return self.parent_node

    def has_child(self):
        return True if len(self.child_node_dict) > 0 else False

    def delete_child(self, child_node):
        node_value = child_node.get_node_value()
        if node_value in self.child_node_dict:
            self.child_node_dict.pop(node_value)

    def get_child_by_value(self, value):
        if value in self.child_node_dict:
            return self.child_node_dict[value]
        else:
            return None

    def add_child_node(self, value, child_node):
        self.child_node_dict[value] = child_node

    def get_node_value(self):
        return self.node_value

    def set_meta_data(self, meta_data):
        self.meta_data = meta_data

    def get_meta_data(self):
        return self.meta_data

    def get_node_status(self):
        return self.node_status

    def set_as_end_node(self):
        self.node_status = DictTrieNode.STATUS_END

    def is_end_node(self):
        if self.node_status == DictTrieNode.STATUS_END:
            return True

        return False

    def set_node_status(self, node_status):
        self.node_status = node_status

    def clear_node_data(self):
        """
        清空数据,删除节点时使用，但不能删除value等数据，因为它可能是某一条路径的中间节点
        :return:
        """
        self.meta_data = None
        self.node_status = DictTrieNode.STATUS_MIDDLE


class DictTrie(object):
    def __init__(self):
        self.root_node = DictTrieNode()

    def insert(self, value_list, meta_data=None):
        """
        添加数据
        :param value_list: list(string)
        :param meta_data: 单条数据的元数据
        :return:
        """
        cur_node = self.root_node

        for i, value in enumerate(value_list):
            child_node = cur_node.get_child_by_value(value)
            if child_node is None:
                child_node = DictTrieNode(value, cur_node)
                cur_node.add_child_node(value, child_node)
            cur_node = child_node

            # 最后一条数据了
            if i == len(value_list) - 1:
                cur_node.set_meta_data(meta_data)
                cur_node.set_as_end_node()

    def __find_node(self, value_list):
        cur_node = self.root_node
        for i, value in enumerate(value_list):
            child_node = cur_node.get_child_by_value(value)
            if child_node is None:
                break
            else:
                cur_node = child_node

            if i == len(value_list) - 1:
                if cur_node.is_end_node():
                    return cur_node

        return None

    def search_node(self, value_list):
        """
        查找节点
        完整匹配
        :param value_list:list(string)
        :return:
        """
        target_node = self.__find_node(value_list)
        if target_node is None:
            return None
        value = target_node.get_node_value()
        meta_data = target_node.get_meta_data()

        return value, meta_data

    def search_path(self, value_list):
        """
        查找路径
        查找部分，取最长的匹配路径
        :param value_list:
        :return:
        """
        candidate_value_list = list()
        target_value_list = None
        target_meta = None

        cur_node = self.root_node
        for i, value in enumerate(value_list):
            child_node = cur_node.get_child_by_value(value)
            if child_node is None:
                break
            else:
                candidate_value_list.append(value)
                if child_node.is_end_node():
                    target_value_list = candidate_value_list[:]
                    target_meta = child_node.get_meta_data()
                cur_node = child_node

        return target_value_list, target_meta

    def search_path_for_all(self, value_list):
        """
        从头开始查找所有符合的数据
        :param value_list:
        :return:
        """
        start = 0
        result_list = list()

        while start < len(value_list):
            search_text = value_list[start:]
            r_result = self.search_path(search_text)
            if r_result is None or r_result[1] is None:
                start += 1
            else:
                find_item = dict()
                find_item['value'] = r_result[0]
                find_item['meta'] = r_result[1]
                find_item['start'] = start
                find_item['end'] = start + len(find_item['value'])
                result_list.append(find_item)

                t_text = find_item['value']
                start += len(t_text)

        return result_list

    def delete(self, value_list):
        """
        删除对应的节点
        :param value_list:
        :return:
        """
        target_node = self.__find_node(value_list)
        if target_node:
            target_node.clear_node_data()
            if not target_node.has_child():
                parent_node = target_node.get_parent_node()
                if parent_node:
                    parent_node.delete_child(target_node)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : dada
# @time   : 2020/7/7 17:00



# class TagDeal():
#     pass
import pickle
from lavda.backexcutor.analysis_interface import DfProcessInterface, RowInterface
from lavda.analysis.da_tag.mapping_models import MappingCollect, Relation_Mapping_collect
from lavda.analysis.da_tag.split_tag import max_split_tag, max_relation_split_tag
from lavda.analysis.da_tag.tag import TextMatch2, deal_sentence_tag, TagContentChoiceCollect
from lavda.analysis.da_tag.tag_config import tag_choice_dic
from lavda.analysis.da_tag.Filed_Services import VerMappingCollect

the_clean_list = ['tag_clean_len','tag_clean_no_chinese_english','tag_clean_weibo_ads','tag_clean_ec_ads','tag_clean_weibo_water']

class TagConfig:
    pass
    # def __init__(self):
    #     self.name = ""



class TagDealSimple(RowInterface,TagConfig):

    def __init__(self, tag_sen_list=None, distance_dict=None,default_value=None):
        super(TagDealSimple, self).__init__(default_value)
        self.tag_sen_list = tag_sen_list
        self.distance_dict = distance_dict

    def load_config_data(self, **config):

        merge_dict_path = config['mappingDictPath']
        check_result = VerMappingCollect(merge_dict_path).ver_status
        assert list(check_result.keys())[0] != 1, "关键词表有错"
        # 合并结果数据
        mapping = MappingCollect(merge_dict_path)
        mapping.add_sen_sheet(self.tag_sen_list)
        mapping.add_length_sheet(self.distance_dict)
        mapping.to_pickle(merge_dict_path.replace("xlsx","pkl"))
        self.mapping = mapping


    def process(self, data, **params):
        merge_content = TagContentChoiceCollect(tag_choice_dic)
        line_column_list = list(data.keys())
        filter_clean_list = list(filter(lambda column: column in the_clean_list, line_column_list))
        filter_clean_list_value = [int(data[value]) for value in filter_clean_list]
        tag_dict = dict()
        for _map in self.mapping:
            # if sum(filter_clean_list_value) > 0:
            #     tag_dict[_map.sheetname] = ""
            # else:
            tag_dict[_map.sheetname] = TextMatch2.one_match_multi_mix(data, _map,merge_content,default_len=7)

        return str(tag_dict)



class ResultTagFormat(DfProcessInterface):
    def __init__(self, intermediate, add_head_text):
        self.intermediate = intermediate
        self.add_head_text = add_head_text

    def load_config_data(self, **config):
        tag_dict_path = config['TagDictPath']
        tag_dict_path = tag_dict_path.replace("xlsx", "pkl")
        tag_dict = open(tag_dict_path, 'rb')
        mapping = pickle.load(tag_dict)
        self.mapping = mapping

    def process(self, data, **params):
        if self.mapping.relation:
            data = max_relation_split_tag(data,self.mapping,self.intermediate,self.add_head_text)
        else:
            data = max_split_tag(data,self.mapping,self.intermediate,self.add_head_text)
        return data



class TagDealRelationSimple(RowInterface,TagConfig):

    def __init__(self, tag_sen_list=None, distance_dict=None, default_value=None):
        super(TagDealRelationSimple, self).__init__((default_value, default_value, default_value))


    def load_config_data(self, **config):

        merge_dict_path = config['mappingDictPath']
        check_result = VerMappingCollect(merge_dict_path).ver_status
        assert list(check_result.keys())[0] != 1, "关键词表有错"
        # 合并结果数据
        mapping = Relation_Mapping_collect(merge_dict_path)
        mapping.to_pickle(merge_dict_path.replace("xlsx","pkl"))
        self.mapping = mapping
        self._ = 0
        for map_i in self.mapping:
            if map_i.sheetname == '品牌' and '品牌(需求)' in map_i.mapping_df.columns:
                self.map_1 = map_i.mapping_df[map_i.mapping_df['品牌(需求)'] == 1]
                self.map_0 = map_i.mapping_df[map_i.mapping_df['品牌(需求)'] == 0]
                self._ = 1

    def process(self, data, **params):
        if self._:
            result = deal_sentence_tag(data, self.mapping, self.map_1, self.map_0)
        else:
            result = deal_sentence_tag(data, self.mapping)
        # result[-1] 存放的是三列需求的数据：关键词包含父级关系，关键词，打到的词，
        # keyword_list = []
        aspect_list = []
        new_aspect_list = []
        keyword_parents_list = []
        for i in result[-1]:
            keyword_parents_list.append(i.split('*')[0])
            # keyword_list.append(i.split('*')[1])
            aspect_list.append(i.split('*')[2])
        for i in aspect_list:
            if ',' in i:
                new_aspect_list.extend(i.split(','))
            else:
                new_aspect_list.append(i)
        return str(result[0]), str(list(set(keyword_parents_list))), str(list(set(new_aspect_list)))



# -*- coding: utf-8 -*-
# ----------------------------------------------
# purpose : 标签extract的业务层
# author : LGD
# create_time : 2021/2/5 10:56
# update_time : 2021/2/5 10:56
# copyright : Lavector
# ----------------------------------------------
import itertools
import pandas as pd

from lavda.util.panda_tool import PandaTool
from lavda.util.mysql_util import CommonLavDbReader
from lavda.util.dict_trie import DictTrie
from lavda.backexcutor.analysis_interface import RowInterface
from lavda.analysis.da_tag.label_dict import PandaLabelDictTool
from lavda.analysis.da_tag.label_keyword import KRuleParser
from lavda.analysis.da_tag.label_util import LabelUtil
from lavda.config.maxda_config import LAV_AI_DB_CONFIG


LABEL_CONTENT_DICT = {
    "redbook": {
        "default": ['content', 'title'],
    },
    "zhihu": {
        "default": ['content', 'title'],
    },
    "xinyang": {
        "default": ['content', 'title'],
    },
    "tiktok": {
        "default": ['content', 'title'],
    },
    "bilibili": {
        "default": ['content', 'title'],
    },
    "boqi": {
        "default": ['content', 'title'],
    },
    "tmall": {
        "品牌": ['content', 'title'],
        "default": ['content'],
    },
    "jd": {
        "品牌": ['content', 'title'],
        "default": ['content'],
    },
    "nosetime": {
        "品牌": ['content', 'title'],
    },

    "default":  ['content']
}


class PurchaseFilter(object):
    """
    购买标签的过虑规则
    """
    def __init__(self):
        df_word = CommonLavDbReader.get_as_df("pro_label_dict", db_config=LAV_AI_DB_CONFIG)
        word_list = df_word.to_dict(orient="records")
        purchase_must_word_set = set([item["term"] for item in word_list if item['category'] == '购买'])
        # purchase_must_word_set = {'做过', '做了', '完成', '术后', '做的', '复购', '买了', '打完', '打了', '买的', '囤的', '囤了', '坚持做', '定期',
        #                           '有用', '第二次', '第三次', '体验了'}

        trie = DictTrie()
        for word in purchase_must_word_set:
            trie.insert(list(word), word)

        self.trie = trie

    def is_skip(self, input_text):
        start = 0
        # result_list = list()

        value_list = list(input_text)
        while start < len(value_list):
            search_text = value_list[start:]
            r_result = self.trie.search_path(search_text)
            if r_result is None or r_result[1] is None:
                start += 1
            else:
                return False
                # find_item = dict()
                # find_item['value'] = r_result[0]
                # find_item['meta'] = r_result[1]
                # find_item['start'] = start
                # find_item['end'] = start + len(find_item['value'])
                # result_list.append(find_item)
                #
                # t_text = find_item['value']
                # start += len(t_text)

        return True


class KeywordBaseExtractor(RowInterface):
    """
    关键词提取、匹配基础功能
    """
    COL_TITLE = 'title'
    COL_CONTENT = 'content'

    def __init__(self, default_value=None):
        super(KeywordBaseExtractor, self).__init__(default_value)
        # TODO 定义打标签功能输出格式

        self._mc = None
        self._map_rules = list()

        self.purchase_filter = None

    @property
    def mc(self):
        return self._mc

    def load_base_config_data(self, **config):
        dict_tool = PandaLabelDictTool()

        label_dict_path = config['mappingDictPath']
        mc = dict_tool.load_all_from_file(label_dict_path)
        for mapping in mc.mappings:
            k_rule = KRuleParser(mapping.id2keyword)
            self._map_rules.append((mapping, k_rule))
        self._mc = mc

        mc_names = [item.label_name for item in mc.mappings]
        if '购买' in mc_names:
            self.purchase_filter = PurchaseFilter()

        return mc

    def load_config_data(self, **config):
        """
        加载词表
        :param config:
        :return:
        """
        self.load_base_config_data(**config)

    def clean_text(self, content):
        clean_data = LabelUtil.clean_text(content)
        return clean_data

    def build_task_params(self, data):
        """
        构建提取任务参数,主要是确定各个标签、各个平台打标所需要对应的字段
        :param data:
        :return:
        """
        channel = data['channel']

        # 1. 提取内容列
        input_task_list = list()
        input_columns_set = set()
        for i, mapping_rule in enumerate(self._map_rules):
            mapping, k_rule = mapping_rule
            label_name = mapping.label_name
            input_columns = None
            if channel in LABEL_CONTENT_DICT:
                if label_name in LABEL_CONTENT_DICT[channel]:
                    input_columns = LABEL_CONTENT_DICT[channel][label_name]
                elif 'default' in LABEL_CONTENT_DICT[channel]:
                    input_columns = LABEL_CONTENT_DICT[channel]['default']
            if input_columns is None:
                input_columns = LABEL_CONTENT_DICT['default']
            input_columns_set.update(input_columns)

            input_task_list.append((mapping, k_rule, input_columns))

        return input_task_list, input_columns_set

    def build_texts(self, data, input_columns_set):
        """
        构建并格式化输入文本,主要是数据清洗并格式化
        :param data: dict
        :param input_columns_set:
        :return:
        """
        input_data_dict = dict()
        for col in input_columns_set:
            col_data = self.clean_text(data[col])
            input_data_dict[col] = col_data

        return input_data_dict

    def parse_keywords(self, data, return_all=False):
        """
        从文本中提取标签关键词
        :param data:
        :param return_all: bool，是否提取所有关键词
        :return:
                tuple(input_data_dict, keyword_result)
                input_data_dict:格式化的内容
                keyword_result: 匹配结果

                eg:
                input_data_dict:{'content': [{'index': 0, 'content': '跟tf80对比，它是有点雾面感，不滋润，比较干，膏体是像蜡笔一样的感觉，个人更喜欢滋润的唇膏', 'start': 0, 'end': 46}]}

                keyword_result(每个元素是一个词表的结果):
                [{'content': [{'aspect': '不滋润', 'start': 16, 'end': 19, 'keyword': '不+滋润|有点干|太干|无敌干|拔干-不拔干', 'kid': 47, 'index': 0},
                             {'aspect': '不滋润', 'start': 16, 'end': 19, 'keyword': '不+滋润|有点干|太干|无敌干|拔干-不拔干', 'kid': 250, 'index': 0},
                             {'aspect': '滋润', 'start': 17, 'end': 19, 'keyword': '保湿|滋润|补水|缓解干燥|不起皮', 'kid': 49, 'index': 0},
                             {'aspect': '喜欢', 'start': 39, 'end': 41, 'keyword': '喜欢|心水|感兴趣', 'kid': 282, 'index': 0},
                             {'aspect': '滋润', 'start': 41, 'end': 43, 'keyword': '滋润-不滋润|不拔干|特别润|不干涩|非常润', 'kid': 50, 'index': 0}
                             ]
                }]
        """

        # 1.构建任务参数
        input_task_list, input_columns_set = self.build_task_params(data)

        # 2.构建所需要的数据格式
        input_data_dict = self.build_texts(data, input_columns_set)

        # 3.实际提取匹配
        keyword_result = [{col: list() for col in input_data_dict} for _ in range(len(input_task_list))]
        for one_col in input_data_dict:
            for i, task_item in enumerate(input_task_list):
                mapping, k_rule, input_columns = task_item
                if one_col not in input_columns:
                    continue

                # TODO 单独给购买添加的一个过虑，临时写法，而且是使用原始数据，而非清洗后的数据
                if mapping.label_name == "购买" and self.purchase_filter is not None:
                    if self.purchase_filter.is_skip(data[one_col]):
                        continue
                if one_col == self.COL_TITLE:
                    one_result_dict = self.parse_from_title(mapping, k_rule, input_data_dict[one_col],
                                                            return_all=return_all, original_data=data)
                elif one_col == self.COL_CONTENT:
                    one_result_dict = self.parse_from_content(mapping, k_rule, input_data_dict[one_col],
                                                              return_all=return_all, original_data=data)
                else:
                    raise Exception("unknown col for parser=" + str(one_col))

                if one_result_dict is None:
                    continue
                keyword_result[i][one_col].extend(one_result_dict)

        # 暂时不返回具体的匹配信息
        input_data_dict = self.format_input_data(input_data_dict)
        return input_data_dict, keyword_result

    def is_skip_match(self, mapping, one_col, data):
        """
        正式使用规则匹配前的一些校验、过虑、检查，以确定是否需要进入下一下该步，
        TODO 这些功能，放这里实在是不优雅
        :return:
        """




    def format_input_data(self, input_data_dict):
        """
        格式化/标准化输入数据，主要是分句时使用
        :param input_data_dict:
        :return:
        """
        return input_data_dict

    def parse_from_text(self, mapping, k_rule, content, return_all=False, original_data=None, rule_param=None):
        total_params = {'content': content,
                        'return_all': return_all}
        if rule_param is not None:
            total_params.update(rule_param)
        # one_result = k_rule.parse(content=content, return_all=return_all)
        one_result = k_rule.parse(**total_params)

        return one_result

    def build_title_rule_param(self, original_data):
        # TODO 业务级约定：jd、tmall title作特殊处理,不限制距离
        if original_data is not None and 'channel' in original_data and original_data['channel'] in {"tmall", 'jd'}:
            rule_param = dict()
            rule_param['plus_max_len'] = 100
            return rule_param
        else:
            return None

    def parse_from_title(self, mapping, k_rule, content, return_all=False, original_data=None):
        """
        从title字段提取关键词
        :param mapping:
        :param content:
        :param k_rule:
        :param return_all:
        :return:
        """
        rule_param = self.build_title_rule_param(original_data)
        one_result = self.parse_from_text(mapping, k_rule, content, return_all, rule_param=rule_param)

        return one_result

    def parse_from_content(self, mapping, k_rule, content, return_all=False, original_data=None):
        """
        从content字段提取关键词
        :param mapping:
        :param content:
        :param k_rule:
        :param return_all:
        :return:
        """
        one_result = self.parse_from_text(mapping, k_rule, content, return_all)
        return one_result

    def process(self, data, **params):
        """

        :param data: dict
                    eg: {'channel': 'nosetime', 'content': '也是秋冬必备的颜色之一', 'title': '纪梵希绝美新色，颜色漂亮呀'}
        :param params:
        :return:
                eg:
                []
        """
        return_all = False
        merge_and_format = False
        if params is not None:
            if 'return_all' in params:
                return_all = params['return_all']
            if 'merge_and_format' in params:
                merge_and_format = params['merge_and_format']

        # result_dict: tuple(dict, list), dict是各字段内容，list是各标签打上的结果
        #    ( {'title': '纪梵希绝美新色，颜色漂亮呀', 'content': '也是秋冬必备的颜色之一'},
        #      [{'title': [{'aspect': '纪梵希', 'start': 0, 'end': 3, 'keyword': '纪梵希|givenchy', 'kid': 391}],  'content': []},
        #       {'title': [], 'content': [{'aspect': '秋冬必备的颜色', 'start': 2, 'end': 9, 'keyword': '颜色+秋冬', 'kid': 289}]}]
        #      )
        #
        result_dict = self.parse_keywords(data, return_all)

        if not merge_and_format:
            return result_dict
        else:
            merge_data = self.merge_and_format(result_dict[-1])
            last_result = self.build_result_format(result_dict[-1], merge_data)
            return last_result

    def merge_and_format(self, keyword_result, input_data_dict=None, default_value=None):
        """
        合并不同字段抽取结果并格式化
        :param keyword_result: list, 来处parse_keywords函数的结果
        :return:
            eg: 两个词表，每一列，表示一个词表
                [(30, 283),
                (30, 67)]
        """
        kid_result = list()
        for one_dict in keyword_result:
            one_kid_list = list()
            # TODO 不同字段关键词结果进行合并
            for col in one_dict:
                one_kid_list.extend(one_dict[col])

            one_kids = list(set([kk_item['kid'] for kk_item in one_kid_list]))
            kid_result.append(one_kids)

        # 直接迪卡尔积扩展
        merge_data = KeywordExtractUtil.expand_and_product(*kid_result, default_value=default_value)

        return merge_data

    def build_result_format(self, keyword_result, merge_data):
        """
        构建输出格式，3个字段
        :param keyword_result:
        :param merge_data:
        :return:
        """
        tag_keyword_set = set()
        tag_aspect_list = list()
        for one_dict in keyword_result:
            # TODO 不同字段关键词结果进行合并
            kid_aspect = dict()
            for col in one_dict:
                col_result = one_dict[col]
                for item in col_result:
                    kid = item['kid']
                    if kid not in kid_aspect:
                        kid_aspect[kid] = set()
                    kid_aspect[kid].add(item['aspect'])

                tag_keyword_set.update([item['keyword'] for item in col_result])
            tag_aspect_list.append(kid_aspect)

        tag_keyword_set = list(tag_keyword_set)
        tag_aspect_set = tag_aspect_list

        return merge_data, tag_keyword_set, tag_aspect_set


def eval_result_list(input_str):
    """
    可能由于excel保存时丢失一些内容,因此要校验
    :param input_str:
    :return:
    """
    tmp_list = list(input_str)
    if tmp_list[-1] == ']':
        eval_result = eval(input_str)
        return eval_result
    pos = len(tmp_list) - 1
    while pos >= 0:
        if tmp_list[pos] == ')':
            break
        pos -= 1

    select_list = tmp_list[:pos+1]
    select_list += ']'
    result_str = "".join(select_list)
    eval_result = eval(result_str)

    # TODO 因为结果数据太多时，一般是有问题的，因此象征性的只取一个结果吧
    if len(eval_result) > 0:
        eval_result = eval_result[: 1]
    return eval_result


class KeywordExtractUtil(object):
    @staticmethod
    def expand_and_format(label_result,
                          mc,
                          tag_result_col="tag_result",
                          tag_aspect_col='tag_aspect',
                          tag_keyword_col="tag_keyword",
                          tmp_index_col="tmp_merge_index",
                          result_prefix="tag_format_"):
        """
        将标签的结果进行格式化展开，以方便分析师分析
        :param label_result: string(结果文件) or dataframe
        :param mc: MappingCollect
        :param tag_result_col: string, 打标签的中间结果字段名
        :param tag_aspect_col: string, 存储标签aspect信息的字段
        :param tag_keyword_col: string, 存储keyword信息的列，这里只是用于删除的，没别的用处
        :param tmp_index_col: string,展开时使用的临时索引字段名
        :param result_prefix: string,展开后结果字段前缀
        :return:
        """
        mappings = mc.mappings
        total_columns = list()
        mapping_col_dict = dict()
        for i, mapping in enumerate(mappings):
            tmp_cols = [result_prefix + col for col in mapping.columns]
            # TODO 加一个aspect列
            tmp_cols.append(result_prefix + mapping.label_name + "(aspect)")
            mapping_col_dict[i] = tmp_cols
            total_columns.extend(tmp_cols)

        df_input = PandaTool.read_more(label_result)
        df_input[tmp_index_col] = range(df_input.shape[0])

        split_lines = df_input.groupby(tmp_index_col)
        # df_sub_list = list()
        line_result = list()
        for split_index, df_line in split_lines:
            label_kid_list = df_line[tag_result_col]
            tmp_label_id_value = label_kid_list.values[0]
            if isinstance(tmp_label_id_value, str):
                label_kid_list = eval_result_list(tmp_label_id_value)
            else:
                label_kid_list = tmp_label_id_value

            label_aspect_df = df_line[tag_aspect_col]
            if isinstance(tmp_label_id_value, str):
                label_aspect_list = eval(label_aspect_df.values[0])
            else:
                label_aspect_list = label_aspect_df.values[0]

            for line_kids in label_kid_list:
                line_keyword_result = list()
                # 处理一行所有的标签信息
                for i, one_kid in enumerate(line_kids):
                    if one_kid is not None:
                        id2relation = mappings[i].id2relation
                        # TODO 添加aspect结果数据
                        keyword_aspect = ",".join(label_aspect_list[i][one_kid])
                        keyword_line = id2relation[one_kid].copy()
                        keyword_line.append(keyword_aspect)
                        # print(i, len(keyword_line), len(id2relation[one_kid]))
                    else:
                        # TODO 空标签用空值填充
                        keyword_line = [None for _ in range(len(mapping_col_dict[i]))]

                    line_keyword_result.extend(keyword_line)
                # TODO 最后一个元素即tmp_index_col作为和原始数据关联使用
                line_keyword_result = line_keyword_result + [split_index]
                line_result.append(line_keyword_result)
        # 纯标签信息
        db_only_label = pd.DataFrame(line_result, columns=total_columns+ [tmp_index_col])

        inner_line = pd.merge(df_input, db_only_label, on=tmp_index_col, how='left')
        df_data = inner_line.drop(columns=[tmp_index_col, tag_result_col, tag_aspect_col, tag_keyword_col])

        return df_data

    @staticmethod
    def old_expand_and_format(label_result,
                          mc,
                          tag_result_col="tag_result",
                          tag_aspect_col='tag_aspect',
                          tag_keyword_col="tag_keyword",
                          tmp_index_col="tmp_merge_index",
                          result_prefix="tag_format_"):
        """
        将标签的结果进行格式化展开，以方便分析师分析
        旧的方法，仿照达达写的，效率有些低，因此废弃了
        :param label_result: string(结果文件) or dataframe
        :param mc: MappingCollect
        :param tag_result_col: string, 打标签的中间结果字段名
        :param tag_aspect_col: string, 存储标签aspect信息的字段
        :param tag_keyword_col: string, 存储keyword信息的列，这里只是用于删除的，没别的用处
        :param tmp_index_col: string,展开时使用的临时索引字段名
        :param result_prefix: string,展开后结果字段前缀
        :return:
        """
        mappings = mc.mappings
        total_columns = list()
        mapping_col_dict = dict()
        for i, mapping in enumerate(mappings):
            tmp_cols = [result_prefix + col for col in mapping.columns]
            # TODO 加一个aspect列
            tmp_cols.append(result_prefix + mapping.label_name + "(aspect)")
            mapping_col_dict[i] = tmp_cols
            total_columns.extend(tmp_cols)

        df_input = PandaTool.read_more(label_result)
        df_input[tmp_index_col] = range(df_input.shape[0])

        split_lines = df_input.groupby(tmp_index_col)
        df_sub_list = list()
        for split_index, df_line in split_lines:
            label_kid_list = df_line[tag_result_col]
            tmp_label_id_value = label_kid_list.values[0]
            if isinstance(tmp_label_id_value, str):
                label_kid_list = eval(tmp_label_id_value)
            else:
                label_kid_list = tmp_label_id_value

            label_aspect_df = df_line[tag_aspect_col]
            if isinstance(tmp_label_id_value, str):
                label_aspect_list = eval(label_aspect_df.values[0])
            else:
                label_aspect_list = label_aspect_df.values[0]

            line_result = list()
            for line_kids in label_kid_list:
                line_keyword_result = list()
                for i, one_kid in enumerate(line_kids):
                    if one_kid is not None:
                        id2relation = mappings[i].id2relation
                        # TODO 添加aspect结果数据
                        keyword_aspect = ",".join(label_aspect_list[i][one_kid])
                        keyword_line = id2relation[one_kid].copy()
                        keyword_line.append(keyword_aspect)
                        # print(i, len(keyword_line), len(id2relation[one_kid]))
                    else:
                        # TODO 空标签用空值填充
                        keyword_line = [None for _ in range(len(mapping_col_dict[i]))]

                    line_keyword_result.extend(keyword_line)
                line_result.append(line_keyword_result)

            db_only_label = pd.DataFrame(line_result, columns=total_columns)
            db_only_label[tmp_index_col] = split_index

            inner_line = pd.merge(df_line, db_only_label, on=tmp_index_col, how='left')
            line = inner_line.drop(columns=[tmp_index_col, tag_result_col, tag_aspect_col, tag_keyword_col])

            df_sub_list.append(line)

        # 合并所有
        df_data = pd.concat(df_sub_list, sort=False)

        return df_data

    @staticmethod
    def expand_and_product(*input_list, default_value=None):
        """
        迪卡尔积（只是添加一些扩展功能）
        :param input_list:
        :param default_value:
        :return:
        """
        for one_list in input_list:
            if len(one_list) == 0:
                one_list.append(default_value)

        data = list(itertools.product(*input_list))
        return data

    @staticmethod
    def distinct_keywords_kid(input_list):
        """
        二维数组去重，业务级的去重
        保留关系内容最全面的
        # TODO 空值用None 表示
        eg:(k1, None, k2), (k1, None, None) 去重后 => (k1, None, k2)
        :param input_list:
        :return:
        """
        if len(input_list) <= 1:
            return input_list

        # 提取非空元素数据
        elem_list = [set() for _ in range(len(input_list))]
        for i, line in enumerate(input_list):
            for j, value in enumerate(line):
                # TODO 空值用None来表示
                if value is None:
                    continue
                # TODO 因为不同行的值可能是唯一的，因此加一个索引前缀，以产生唯一的值
                u_elem = str(j) + "-" + str(value)
                elem_list[i].add(u_elem)

        # 合并交集部分
        flag_list = [True for _ in range(len(input_list))]
        for i in range(len(input_list) - 1):
            for j in range(i+1, len(input_list)):
                if not flag_list[i] or not flag_list[j]:
                    continue
                ei = elem_list[i]
                ei_count = len(ei)
                if ei_count <= 0:
                    flag_list[i] = False
                    continue

                ej = elem_list[j]
                ej_count = len(ej)
                if ej_count <= 0:
                    flag_list[j] = False
                    continue

                e_tmp_set = ei.intersection(ej)
                e_tmp_count = len(e_tmp_set)

                if e_tmp_count == ei_count:
                    flag_list[i] = False
                elif e_tmp_count == ej_count:
                    flag_list[j] = False
                else:
                    continue

        # 返回最终结果
        result_list = [one_item for i, one_item in enumerate(input_list) if flag_list[i]]

        return result_list


def segment_text(text, split_char_set):
    """
    分句
    :param text:
    :param split_char_set:
    :return:
    """
    result_list = list()

    tmp_sub_list = list()
    text_len = len(text)
    sub_text_start = 0
    cur_char_pos = 0

    for i, ch in enumerate(text):

        # 检查到分隔符
        if ch in split_char_set:
            # TODO 分隔符也是需要保留的
            tmp_sub_list.append(ch)
            sub_text = "".join(tmp_sub_list)
            text_index = len(result_list)
            sub_text_term = {'index': text_index, 'content': sub_text, "start": sub_text_start, 'end': sub_text_start + len(sub_text)}
            result_list.append(sub_text_term)

            sub_text_start = sub_text_term['end']
            tmp_sub_list = list()
            continue
        else:
            tmp_sub_list.append(ch)

        # 最后一个字符
        if text_len - 1 == i:
            if tmp_sub_list:
                sub_text = "".join(tmp_sub_list)
                text_index = len(result_list)
                sub_text_term = {'index': text_index, 'content': sub_text, "start": sub_text_start,
                                 'end': sub_text_start + len(sub_text)}
                result_list.append(sub_text_term)

        cur_char_pos += len(ch)

    return result_list


class KeywordSegmentExtractor(KeywordBaseExtractor):
    """
    基于分句的关键词提取、匹配方法
    """
    def __init__(self, default_value=None):
        super(KeywordSegmentExtractor, self).__init__(default_value)
        # 分句打标功能

        self.split_char = {'.', '。', '!', '！', '\n'}

        # 品牌、消费者需要标签向后扩展
        self.brand_exclude_kids = set()
        self.brand_index = None
        self.need_index = None
        self.expand_need = False

    def load_config_data(self, **config):
        """
        加载词表
        :param config:
        :return:
        """
        mc = self.load_base_config_data(**config)

        # TODO 特殊处理
        flag_col = None
        for i, mapping in enumerate(mc.mappings):
            name = mapping.label_name
            if name == '品牌':

                columns = mapping.columns
                id_col = None
                for col in columns:
                    if '(flag)' in col:
                        flag_col = col
                    if '(id)' in col:
                        id_col = col
                if flag_col is not None:
                    id2line = mapping.id2line
                    for one_line in list(id2line.values()):
                        line_kid = int(one_line[id_col])
                        if int(one_line[flag_col]) == 0:
                            self.brand_exclude_kids.add(line_kid)

                self.brand_index = i
            if name == '消费者需求':
                self.need_index = i

            if flag_col is not None and self.need_index is not None and self.brand_index is not None:
                self.expand_need = True

    def build_texts(self, data, input_columns_set):
        input_data_dict = dict()
        for col in input_columns_set:
            # TODO 要先清洗，再分句，否则分句的结果，不容易对应到主句(主要是指aspect的位置信息)
            # TODO 清洗时，不能将分隔符给清洗掉
            col_data = self.clean_text(data[col])

            col_data = segment_text(col_data, self.split_char)

            input_data_dict[col] = col_data

        return input_data_dict

    def format_input_data(self, input_data_dict):
        """
        分句的内容，再合成一个完整的句子
        :param input_data_dict:
        :return:
        """
        format_data = dict()
        for col in input_data_dict:
            sub_text = "".join(item['content'] for item in input_data_dict[col])
            format_data[col] = sub_text

        return format_data

    def parse_by_segment(self, mapping, k_rule, content, return_all=False, original_data=None, rule_param=None):
        """

        :param mapping:
        :param k_rule:
        :param content: segment_text 拆分后的结果
        :param return_all:
        :return:
        """
        result_list = list()
        for text_item in content:
            sub_content = text_item['content']

            one_result = self.parse_from_text(mapping, k_rule, sub_content, return_all=return_all, original_data=original_data, rule_param=rule_param)

            # TODO 添加分句索引字段，同时将aspect的分句位置替换为在整句的位置
            for r_item in one_result:
                r_item["index"] = text_item['index']
                r_item["start"] = text_item['start'] + r_item['start']
                r_item["end"] = text_item['start'] + r_item['end']
            result_list.extend(one_result)

        return result_list

    def parse_from_content(self, mapping, k_rule, content, return_all=False, original_data=None):
        result_list = self.parse_by_segment(mapping, k_rule, content, return_all=return_all, original_data=original_data)

        return result_list

    def parse_from_title(self, mapping, k_rule, content, return_all=False, original_data=None):
        rule_param = self.build_title_rule_param(original_data)

        result_list = self.parse_by_segment(mapping, k_rule, content, return_all=return_all,
                                            original_data=original_data, rule_param=rule_param)

        return result_list

    def build_result_format(self, keyword_result, merge_data):
        """
        构建输出格式，3个字段
        :param keyword_result:
        :param merge_data:
        :return:
        """
        tag_keyword_set = set()
        tag_aspect_list = list()
        for one_dict in keyword_result:
            # TODO 不同字段关键词结果进行合并
            kid_aspect = dict()
            for col in one_dict:
                col_result = one_dict[col]
                for item in col_result:
                    kid = item['kid']
                    if kid not in kid_aspect:
                        kid_aspect[kid] = set()
                    kid_aspect[kid].add(item['aspect'])

                tag_keyword_set.update([item['keyword'] for item in col_result])
            tag_aspect_list.append(kid_aspect)

        tag_keyword_set = list(tag_keyword_set)
        tag_aspect_set = tag_aspect_list


        # TODO 排除不需要的品牌aspect
        if len(self.brand_exclude_kids) > 0:
            t_brand_dict = dict()
            for o_kid in tag_aspect_set[self.brand_index]:
                if o_kid in self.brand_exclude_kids:
                    continue
                t_brand_dict[o_kid] = tag_aspect_set[self.brand_index][o_kid]
            tag_aspect_set[self.brand_index] = t_brand_dict

        return merge_data, tag_keyword_set, tag_aspect_set

    def merge_and_format(self, keyword_result, input_data_dict=None, default_value=None):
        """
        合并不同字段抽取结果并格式化
        :param keyword_result: list, 来处parse_keywords函数的结果
        :param input_data_dict: list, 来处parse_keywords函数的结果
        :param default_value: list,
        :return:
        """
        # 1.不分句字段的结果
        # TODO 所有字段都分句，因此不存在不分句的字段
        # other_kid_list = list()
        # for one_dict in keyword_result:
        #     one_kid_list = list()
        #     for col in one_dict:
        #         if col == self.COL_CONTENT:
        #             continue
        #         one_kid_list.extend(one_dict[col])
        #
        #     one_kids = list(set([kk_item['kid'] for kk_item in one_kid_list]))
        #     other_kid_list.append(one_kids)
        #
        # other_result = KeywordExtractUtil.expand_and_product(*other_kid_list, default_value=default_value)
        other_result = list()

        # 2.分句字段的结果
        # 2.1先提取分句索引
        segment_index = dict()
        for i, one_dict in enumerate(keyword_result):

            for one_col in one_dict:
                for item in one_dict[one_col]:
                    one_index = item['index']
                    seg_index = str(one_col) + '-' + str(one_index)
                    if seg_index not in segment_index:
                        segment_index[seg_index] = dict()
                    if i not in segment_index[seg_index]:
                        segment_index[seg_index][i] = set()
                    segment_index[seg_index][i].add(item['kid'])

        # TODO 品牌-消费者向下扩展，暂时只扩展content字段
        if self.expand_need:
            tmp_expand_brand_and_need(self.COL_CONTENT, segment_index, self.brand_index, self.need_index, self.brand_exclude_kids)
            # TODO 清除残存的不需要的品牌标签结果
            for one_index in segment_index:
                if self.brand_index in segment_index[one_index]:
                    tmp_b_index = set()
                    for one_brand_kid in segment_index[one_index][self.brand_index]:
                        if one_brand_kid in self.brand_exclude_kids:
                            continue
                        tmp_b_index.add(one_brand_kid)
                    segment_index[one_index][self.brand_index] = tmp_b_index
        # 2.2 分别提取各分句结果
        segment_result = list()
        # segment_index: 分句索引 => 标签序列索引 => list(kid)
        for one_index in segment_index:
            segment_kid_list = list()
            for i, one_dict in enumerate(keyword_result):
                if i not in segment_index[one_index] or len(segment_index[one_index][i]) <= 0:
                    # TODO 空值用None表示
                    segment_kid_list.append(list())
                else:
                    segment_kid_list.append(segment_index[one_index][i])

            one_segment_result = KeywordExtractUtil.expand_and_product(*segment_kid_list, default_value=default_value)
            segment_result.extend(one_segment_result)

        all_result = other_result + segment_result

        # 3.去重
        # eg:(k1, None, k2), (k1, None, None) 去重后 => (k1, None, k2)
        all_result = KeywordExtractUtil.distinct_keywords_kid(all_result)

        return all_result


def tmp_expand_brand_and_need(target_col, segment_index, brand_index, need_index, brand_exclude_kids):
    """
    临时处理功能，用于临时扩展品牌-消费者需求关系组合
    :return:
    """

    def split_brand(input_brand_sets):
        t_include_set = set()
        t_exclude_set = set()
        for t_brand_kid in input_brand_sets:
            if t_brand_kid in brand_exclude_kids:
                t_exclude_set.add(t_brand_kid)
            else:
                t_include_set.add(t_brand_kid)
        return t_include_set, t_exclude_set

    max_expand_count = 7

    seg_index_list = list(segment_index.keys())
    # TODO 只选择指定的列的结果
    seg_index_list = [item for item in seg_index_list if target_col in item]
    # 分句索引按顺序排列，从开头到时结尾
    seg_index_list = sorted(seg_index_list, key=lambda x: x)

    start = 0
    while start < len(seg_index_list):
        cur_seg_index = seg_index_list[start]
        cur_brand_sets = segment_index[cur_seg_index][brand_index] if brand_index in segment_index[cur_seg_index] else set()
        cur_include_set, cur_exclude_set = split_brand(cur_brand_sets)
        # segment_index[cur_seg_index][brand_index] = cur_include_set
        # TODO 为了不至于扩展太多数据，因此随机选一个进行扩展，其他不选择
        if len(cur_include_set) <= 0:
            start += 1
            continue
        cur_expand_brand_kid = list(cur_include_set)[0]

        next_index = start + 1
        cur_expand_count = 0
        # TODO 防止最后一句陷入死循环
        if next_index >= len(seg_index_list):
            break
        while next_index < len(seg_index_list) and cur_expand_count <= max_expand_count:
            next_seg_index = seg_index_list[next_index]
            next_brand_sets = segment_index[next_seg_index][brand_index] if brand_index in segment_index[next_seg_index] else set()
            next_need_sets = segment_index[next_seg_index][need_index] if need_index in segment_index[next_seg_index] else set()

            next_include_set, next_exclude_set = split_brand(next_brand_sets)
            # segment_index[next_seg_index][brand_index] = next_include_set

            # TODO 到达下一个合法的品牌标签，要退出
            if len(next_include_set) > 0:
                start = next_index
                break
            if len(next_exclude_set) > 0:
                start = next_index + 1
                break
            # 当前句没有消费者需求标签
            if len(next_need_sets) <= 0:
                next_index += 1
                start = next_index
                continue

            # TODO 当前不包含任何品牌标签，只包含消费者需求标签，因此可以扩展，即将上面的brand kid 加入到当前分句中
            if brand_index not in segment_index[next_seg_index]:
                segment_index[next_seg_index][brand_index] = set()
            #
            segment_index[next_seg_index][brand_index].add(cur_expand_brand_kid)
            cur_expand_count += 1
            next_index += 1
            start = next_index


if __name__ == "__main__":
    # 1.测试分句
    text = "中国来.肿。脸.四月份轸!！wolwe"
    split_char = {'.', '。', '!', '！', '\n'}
    # result = segment_text(text, split_char)

    # 2.测试去重
    data = [(1, None, None), (None, None, 2), (1, None, 2), (18, 19, None)]
    result = KeywordExtractUtil.distinct_keywords_kid(data)
    print(result)


